from collections import Counter

from structlog.stdlib import get_logger

from ice.agents.base import Agent
from ice.agents.base import Stop
from ice.apis.openai import openai_complete

log = get_logger()


class OpenAIReasoningAgent(Agent):
    def __init__(
        self,
        model: str = "text-davinci-002",
        temperature: float = 0.0,
        top_p: float = 1.0,
        num_workers: int = 1,
        reasoning_prefix: str = "Reasoning: ",
    ):
        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.num_workers = num_workers
        self.reasoning_prefix = reasoning_prefix

    def _answer_prefix(self, prompt: str) -> str:
        return prompt.split("\n")[-1]

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ) -> str:
        answer_prefix = self._answer_prefix(prompt)

        # Generate the prompt for the reasoning task
        reasoning_prompt = self._generate_reasoning_prompt(prompt)

        # Request multiple completions from the API
        response = await self._request_completions(reasoning_prompt)

        # Parse the responses and aggregate the answers and reasonings
        answers, reasonings = await self._parse_and_aggregate_responses(
            reasoning_prompt, response, answer_prefix, stop=stop
        )

        # Return the most common answer (TODO: Smarter aggregation)
        return answers.most_common(1)[0][0]

    async def predict(self, *, context, default="", verbose=False) -> dict[str, float]:
        raise NotImplementedError

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: str | None = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], str | None]:
        # Generate the prompt for the reasoning task
        answer_prefix = self._answer_prefix(prompt)
        reasoning_prompt = self._generate_reasoning_prompt(prompt)

        # Request multiple completions from the API
        response = await self._request_completions(reasoning_prompt)

        # Parse the responses and aggregate the answers and reasonings
        answers, reasonings = await self._parse_and_aggregate_responses(
            reasoning_prompt, response, answer_prefix, choices=choices
        )

        # Return a dict [str, float] and the joined reasonings
        return self._format_result(answers, reasonings)

    def _generate_reasoning_prompt(self, prompt: str) -> str:
        answer_prefix = self._answer_prefix(prompt)

        # Replace the answer prefix with the reasoning prefix
        prompt = prompt[: -len(answer_prefix)] + self.reasoning_prefix
        return prompt

    async def _request_completions(self, prompt: str) -> dict:
        # Adjust the temperature based on the number of workers
        temperature = (
            self.temperature
            if self.num_workers == 1
            else min(self.temperature + 0.4, 1.0)
        )

        # Request multiple completions from the API
        response = await openai_complete(
            prompt,
            stop=None,
            model=self.model,
            temperature=temperature,
            top_p=self.top_p,
            n=self.num_workers,
        )
        return response

    async def _parse_and_aggregate_responses(
        self,
        prompt: str,
        response: dict,
        answer_prefix: str,
        stop: Stop = None,
        choices: tuple[str, ...] | None = None,
    ) -> tuple[Counter[str], list[str]]:
        # Extract the response texts
        response_texts = [choice["text"] for choice in response["choices"]]

        # Parse the responses and aggregate the answers and reasonings
        answers: Counter[str] = Counter()
        reasonings: list[str] = []
        for (i, response_text) in enumerate(response_texts):
            # Check if the response contains the answer prefix
            if answer_prefix not in response_text:
                # If not, request an explicit answer from the API
                response_text = await self._request_explicit_answer(
                    prompt, response_text, answer_prefix, stop=stop
                )

            # Parse the answer and the reasoning from the response
            answer, reasoning = self._parse_answer_and_reasoning(
                response_text, answer_prefix, stop=stop
            )

            # Update the answer counts and the reasoning list
            if choices is not None:
                for choice in choices:
                    if answer.strip().startswith(choice.strip()):
                        answers[choice] += 1
            else:
                answers[answer] += 1

            reasonings.append(reasoning)

        return answers, reasonings

    async def _request_explicit_answer(
        self, prompt: str, response_text: str, answer_prefix: str, stop: Stop = None
    ) -> str:
        # Generate a follow-up prompt with the answer prefix
        followup_prompt = f"{prompt}{response_text}\n\n{answer_prefix}"

        # Request a single completion from the API
        followup_response = await openai_complete(
            followup_prompt,
            stop=stop,
            model=self.model,
            temperature=self.temperature,
            top_p=self.top_p,
            n=1,
        )

        # Extract the follow-up response text
        followup_response_text = self._enforce_stop(
            followup_response["choices"][0]["text"], stop
        )

        # Append the follow-up response text to the original response text
        response_text += f"\n\n{answer_prefix}{followup_response_text}"

        return response_text

    def _enforce_stop(self, response_text: str, stop: Stop) -> str:
        if stop is None:
            return response_text.strip()
        return response_text.strip().split("".join(stop))[0]

    def _parse_answer_and_reasoning(
        self, response_text: str, answer_prefix: str, stop: Stop
    ) -> tuple[str, str]:
        # Split the response text by the answer prefix
        response_parts = response_text.split(answer_prefix, maxsplit=1)

        # Check that the response has two parts
        if len(response_parts) != 2:
            log.warning(f"Unexpected response format: {response_text}")
            raise ValueError(f"Unexpected response format: {response_text}")

        # Strip the whitespace from the parts
        reasoning, answer = map(str.strip, response_parts)

        # Check that the reasoning and the answer are not empty
        if not reasoning or not answer:
            log.warning(f"Empty reasoning or answer: {response_text}")

        answer = self._enforce_stop(answer, stop)

        return answer, reasoning

    def _format_result(
        self, answers: Counter[str], reasonings: list[str]
    ) -> tuple[dict[str, float], str]:

        # Join the reasonings with counts
        joined_reasonings = self._join_texts_with_counts(reasonings)

        # Convert the answers counter to a dictionary of probabilities
        total = sum(answers.values())
        answer_probs = {k: v / total for k, v in answers.items()}

        # Sort the dictionary by descending probability
        answer_probs = dict(
            sorted(answer_probs.items(), key=lambda x: x[1], reverse=True)
        )
        return answer_probs, joined_reasonings

    def _join_texts_with_counts(self, texts: list[str]) -> str:
        """
        Given a list like ["bla", "blubb", "bla", "foo", "blubb"], return
        an answer like "(3) bla\n\n(2) blubb\n\n(1)foo".
        """
        counts = Counter(texts)
        return "\n\n".join(
            f"({count}) {text}" for (text, count) in counts.most_common()
        )



import shelve

from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from ice.agents.base import Agent
from ice.agents.base import Stop


class NotApprovedException(Exception):
    pass


def quoted(s: str):
    return "\n".join(f"> {line}" for line in s.split("\n"))


def is_yes(s: str):
    return s.strip().lower().startswith("y")


class ApprovalAgent(Agent):
    """
    Checks with base agent if action is ok before executing it; if not,
    raises Exception
    """

    def __init__(
        self,
        base_agent: Agent,
        approval_agent: Agent,
        approval_cache_path: Optional[Path] = None,
    ):
        self.base_agent = base_agent
        self.approval_agent = approval_agent
        self.approval_cache_path = approval_cache_path

    @contextmanager
    def approval_cache(self):
        if self.approval_cache_path:
            with shelve.open(self.approval_cache_path.as_posix()) as cache:
                yield cache
        else:
            yield {}

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ):
        completion = await self.base_agent.complete(
            prompt=prompt,
            stop=stop,
            verbose=verbose,
            default=default,
            max_tokens=max_tokens,
        )
        await self._check(prompt=prompt, candidate=completion)
        return completion

    async def relevance(self, *, question, context, verbose=False, default=None):
        score = await self.base_agent.relevance(
            question=question, context=context, verbose=verbose, default=default
        )
        await self._check(
            prompt=f"Score the relevance of the context {context}\n to the question {question}",
            candidate=str(score),
        )
        return score

    async def _check(self, prompt: str, candidate: str):

        approval_prompt = f"""Evaluate whether the following output is correct.

Input:
{quoted(prompt)}

Output:
{quoted(candidate)}

Is this output correct (y/n)?"""

        with self.approval_cache() as cache:
            if approval_prompt in cache:
                assert is_yes(str(cache[approval_prompt]))
                return

        approval_action = await self.approval_agent.complete(prompt=approval_prompt)

        if not is_yes(approval_action):
            raise NotApprovedException

        with self.approval_cache() as cache:
            cache[approval_prompt] = approval_action



from collections.abc import Callable
from collections.abc import Sequence
from functools import partial
from itertools import chain

from ice.apis.openai import TooLongRequestError
from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import load_papers
from ice.paper import Paper
from ice.recipe import recipe
from ice.recipes.consort_flow.baseline_elicit_answer import answer_like_elicit_qa
from ice.recipes.consort_flow.generate_questions import adherence_questions_and_answers
from ice.recipes.consort_flow.generate_questions import arms_questions_and_answers
from ice.recipes.consort_flow.generate_questions import (
    experiments_questions_and_answers,
)
from ice.recipes.consort_flow.types import ConsortFlow
from ice.recipes.experiments_and_arms.golds import get_ea_gs
from ice.recipes.meta.eval_paper_qa.common_baselines import (
    preselected_few_shot_qa_baseline,
)
from ice.recipes.meta.eval_paper_qa.quick_list import quick_list
from ice.recipes.meta.eval_paper_qa.types import PaperQaAnswer
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard
from ice.recipes.meta.eval_paper_qa.utils import convert_demonstration_example
from ice.recipes.program_search.nodes.decontext.decontextualize import paper_decontext
from ice.recipes.program_search.nodes.prune.prune import prune
from ice.recipes.program_search.nodes.prune.prune import prune_with_reasoning
from ice.recipes.program_search.nodes.select.select import (
    filter_by_perplexity_threshold,
)
from ice.recipes.program_search.nodes.select.select import (
    select_results_using_elicit_prompt,
)
from ice.recipes.program_search.nodes.select.select import (
    select_using_elicit_prompt_few_shot,
)
from ice.recipes.program_search.types import remove_lowest_perplexity


def experiments_few_shot_demonstration(
    document_id: str, consolidate: bool = False
) -> Sequence[PaperQaGoldStandard]:
    papers = [
        p
        for p in load_papers(split="validation", question_short_name="experiments_arms")
        if p.document_id != document_id
    ]
    gss = [get_ea_gs(p.document_id) for p in papers]
    paper_gs = list(
        chain(
            *[
                experiments_questions_and_answers(paper, gs, consolidate=consolidate)
                for gs, paper in zip(gss, papers)
                if gs
            ]
        )
    )
    return paper_gs


def arms_few_shot_demonstration(
    document_id: str, consolidate: bool = False
) -> Sequence[PaperQaGoldStandard]:
    papers = [
        p
        for p in load_papers(split="validation", question_short_name="experiments_arms")
        if p.document_id != document_id
    ]
    gss = [get_ea_gs(p.document_id) for p in papers]
    paper_gs = list(
        [
            arms_questions_and_answers(paper, gs, consolidate=consolidate)
            for gs, paper in zip(gss, papers)
            if gs
        ]
    )
    used_gs = []
    for paper in paper_gs:
        for gs in paper:
            # 1 per paper
            used_gs.append(gs)
            break

    return used_gs


def adherence_few_shot_demonstration(
    document_id: str, consolidate: bool = False
) -> Sequence[PaperQaGoldStandard]:
    papers = [
        p
        for p in load_papers(split="validation", question_short_name="consort_flow_v2")
        if p.document_id != document_id
    ]
    gss = [
        get_gold_standard(
            document_id=p.document_id,
            question_short_name="consort_flow_v2",
            model_type=ConsortFlow,
        )
        for p in papers
    ]
    paper_gs = list(
        [
            adherence_questions_and_answers(paper, gs)
            for gs, paper in zip(gss, papers)
            if gs
        ]
    )
    used_gs = []
    for paper in paper_gs:
        for gs in paper:
            # 1 per paper
            used_gs.append(gs)
            break

    return used_gs


def _to_paragraphs(paper: Paper) -> Sequence[str]:
    return [str(p) for p in paper.nonempty_paragraphs()]


def _to_sentences(paper: Paper) -> Sequence[str]:
    return [s for s in paper.sentences() if s]


async def cheating_few_shot_qa_experiments_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=experiments_few_shot_demonstration,
    )


async def cheating_few_shot_qa_arms_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=arms_few_shot_demonstration,
    )


async def cheating_few_shot_qa_experiments_paragraph_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=experiments_few_shot_demonstration,
        paper_division_func=_to_paragraphs,
    )


async def cheating_few_shot_qa_arms_paragraph_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=arms_few_shot_demonstration,
        paper_division_func=_to_paragraphs,
    )


async def cheating_few_shot_qa_experiments_paragraph_reasoning_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=experiments_few_shot_demonstration,
        paper_division_func=_to_paragraphs,
        reasoning=True,
    )


async def cheating_few_shot_qa_arms_paragraph_reasoning_baseline(
    paper: Paper, question: str, gold_support=None
):
    return await preselected_few_shot_qa_baseline(
        paper,
        question,
        gold_support,
        enumerate_answer=True,
        few_shot_demonstration_func=arms_few_shot_demonstration,
        paper_division_func=_to_paragraphs,
        reasoning=True,
    )


async def elicit_baseline_into_answer(paper: Paper, question: str, gold_support=None):
    gold_support  # unused
    texts = _to_paragraphs(paper)
    texts_with_perplexities = await select_results_using_elicit_prompt(
        question=question, texts=texts
    )
    # ~ .78 recall, .10 precision @ threshold=1.07
    selections = filter_by_perplexity_threshold(texts_with_perplexities, threshold=1.16)

    while selections:
        try:
            relevant_str = "\n\n".join([s[0] for s in selections])
            answer = await answer_like_elicit_qa(
                question=question, passage=relevant_str
            )
            answer_as_list = await quick_list(question=question, answer=answer)
            selection_set = set([s[0] for s in selections])
            return PaperQaAnswer(
                answer=answer_as_list,
                support_candidates=texts,
                support_labels=[text in selection_set for text in texts],
                support_scores=[t[1] for t in texts_with_perplexities],
            )
        except TooLongRequestError:
            selections = remove_lowest_perplexity(selections)
    return PaperQaAnswer(
        answer=["The question is not answered in the text."],
        support_candidates=texts,
        support_labels=[False for text in texts],
        support_scores=[t[1] for t in texts_with_perplexities],
    )


async def _decontext_selections(paper: Paper, selections: Sequence[tuple[str, float]]):
    decontexted = await paper_decontext(paper)
    decontexted_paras = [str(p) for p in decontexted.paragraphs]
    original_paras = [str(p) for p in paper.paragraphs]
    assert len(decontexted_paras) == len(original_paras)
    old_to_new = {old: new for old, new in zip(original_paras, decontexted_paras)}
    return [(old_to_new[s[0]], s[1]) for s in selections]


async def _all_options(
    paper: Paper,
    question: str,
    gold_support: None,
    few_shot_demonstration_func: Callable[[str, bool], Sequence[PaperQaGoldStandard]],
    prune_to_max: int | None = None,
    do_prune_reasoning: bool = False,
    do_few_shot_selection: bool = False,
    do_decontext_at_answer: bool = False,
    do_demonstration_answer: bool = False,
    do_demonstration_reasoning: bool = False,
    do_return_list: bool = True,
    cls_threshold: float = 1.06,
):
    gold_support  # unused
    texts = _to_paragraphs(paper)
    supporting_examples = few_shot_demonstration_func(paper.document_id, True)
    paragraph_supporting_examples = [
        await convert_demonstration_example(example, _to_paragraphs)
        for example in supporting_examples
    ]
    if do_few_shot_selection:
        texts_with_perplexities = await select_using_elicit_prompt_few_shot(
            question=question, texts=texts, examples=paragraph_supporting_examples
        )
        # Few-shot prompt needs a lower threshold
        selections = filter_by_perplexity_threshold(
            texts_with_perplexities,
            threshold=cls_threshold,
        )
    else:
        texts_with_perplexities = await select_results_using_elicit_prompt(
            question=question, texts=texts
        )
        selections = filter_by_perplexity_threshold(
            texts_with_perplexities,
            threshold=cls_threshold,
        )

    if prune_to_max:
        if do_prune_reasoning:
            pruned_selections = await prune_with_reasoning(
                question=question,
                texts_with_perplexities=selections,
                max_to_keep=prune_to_max,
            )
        else:
            pruned_selections = await prune(
                question=question,
                texts=[t[0] for t in selections],
                max_to_keep=prune_to_max,
            )

        selections = [t for t in selections if t[0] in pruned_selections]

    if do_decontext_at_answer:
        selections = await _decontext_selections(paper=paper, selections=selections)

    while selections:
        try:
            relevant_str = "\n\n".join([s[0] for s in selections])
            if not do_demonstration_answer:
                answer: str | Sequence[str] = await answer_like_elicit_qa(
                    question=question, passage=relevant_str
                )
                if do_return_list:
                    answer = await quick_list(question=question, answer=answer)  # type: ignore[arg-type]
            else:
                subrecipe_answer = await preselected_few_shot_qa_baseline(
                    paper=paper,
                    question=question,
                    gold_support=gold_support,
                    enumerate_answer=do_return_list,
                    few_shot_demonstration_func=lambda doc_id: few_shot_demonstration_func(
                        doc_id, not do_return_list
                    ),
                    selections=[s[0] for s in selections],
                    paper_division_func=_to_paragraphs,
                    reasoning=do_demonstration_reasoning,
                )
                answer = subrecipe_answer.answer
            selection_set = set([s[0] for s in selections])
            # support_labels = [text in selection_set for text in texts]
            # if not any(support_labels):
            #     breakpoint()
            return PaperQaAnswer(
                answer=answer,
                support_candidates=texts,
                support_labels=[text in selection_set for text in texts],
                support_scores=[t[1] for t in texts_with_perplexities],
            )
        except TooLongRequestError:
            selections = remove_lowest_perplexity(selections)
    return PaperQaAnswer(
        answer=["The question is not answered in the text."]
        if do_return_list
        else "The question is not answered in the text.",
        support_candidates=texts,
        support_labels=[False for text in texts],
        support_scores=[t[1] for t in texts_with_perplexities],
    )


zero_shot_arms_into_answer = partial(
    _all_options, few_shot_demonstration_func=arms_few_shot_demonstration
)
few_shot_arms_into_answer = partial(
    _all_options,
    do_few_shot_selection=True,
    few_shot_demonstration_func=arms_few_shot_demonstration,
)
elicit_prune_arms_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=arms_few_shot_demonstration,
    prune_to_max=7,
    do_prune_reasoning=True,
)
few_shot_arms_prune_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=arms_few_shot_demonstration,
    prune_to_max=7,
    do_few_shot_selection=True,
    do_prune_reasoning=True,
)
zero_shot_arms_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=arms_few_shot_demonstration,
    do_demonstration_answer=True,
)
zero_shot_arms_decontext_then_answer = partial(
    _all_options,
    few_shot_demonstration_func=arms_few_shot_demonstration,
    do_decontext_at_answer=True,
)
zero_shot_arms_decontext_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=arms_few_shot_demonstration,
    do_demonstration_answer=True,
    do_decontext_at_answer=True,
)

zero_shot_experiments_into_answer = partial(
    _all_options, few_shot_demonstration_func=experiments_few_shot_demonstration
)
few_shot_experiments_into_answer = partial(
    _all_options,
    do_few_shot_selection=True,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
)
elicit_prune_experiments_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    prune_to_max=7,
    do_prune_reasoning=True,
)
few_shot_experiments_prune_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    prune_to_max=7,
    do_few_shot_selection=True,
    do_prune_reasoning=True,
)
zero_shot_experiments_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    do_demonstration_answer=True,
)
zero_shot_experiments_few_shot_answer_with_reasoning = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    do_demonstration_answer=True,
    do_demonstration_reasoning=True,
)


zero_shot_experiments_decontext_then_answer = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    do_decontext_at_answer=True,
)
zero_shot_experiments_decontext_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=experiments_few_shot_demonstration,
    do_demonstration_answer=True,
    do_decontext_at_answer=True,
)


zero_shot_adherence_into_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    do_return_list=False,
)
few_shot_adherence_into_answer = partial(
    _all_options,
    do_few_shot_selection=True,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    do_return_list=False,
)
elicit_prune_adherence_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    prune_to_max=7,
    do_prune_reasoning=True,
    do_return_list=False,
)
few_shot_adherence_prune_reasoning_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    prune_to_max=7,
    do_few_shot_selection=True,
    do_prune_reasoning=True,
    do_return_list=False,
)
zero_shot_adherence_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    do_demonstration_answer=True,
    do_return_list=False,
)
zero_shot_adherence_decontext_then_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    do_decontext_at_answer=True,
    do_return_list=False,
)
zero_shot_adherence_decontext_few_shot_answer = partial(
    _all_options,
    few_shot_demonstration_func=adherence_few_shot_demonstration,
    do_demonstration_answer=True,
    do_decontext_at_answer=True,
    do_return_list=False,
)


async def elicit_baseline_prune_then_answer(
    paper: Paper, question: str, gold_support=None
):
    gold_support  # unused
    texts = _to_paragraphs(paper)
    selections = await select_results_using_elicit_prompt(
        question=question, texts=texts
    )
    texts_with_perplexities = await select_results_using_elicit_prompt(
        question=question, texts=texts
    )
    # ~ .78 recall, .10 precision @ threshold=1.07
    selections = filter_by_perplexity_threshold(texts_with_perplexities, threshold=1.16)
    while selections:
        try:
            pruned = await prune(
                question=question,
                texts=[s[0] for s in selections],
                max_to_keep=len(selections) // 2 if len(selections) > 10 else 5,
            )
            relevant_str = "\n\n".join(pruned)
            answer = await answer_like_elicit_qa(
                question=question, passage=relevant_str
            )
            answer_as_list = await quick_list(question=question, answer=answer)
            selection_set = set(pruned)
            return PaperQaAnswer(
                answer=answer_as_list,
                support_candidates=texts,
                support_labels=[text in selection_set for text in texts],
                support_scores=[t[1] for t in texts_with_perplexities],
            )
        except TooLongRequestError:
            selections = remove_lowest_perplexity(selections)
    return PaperQaAnswer(
        answer=["The question is not answered in the text."],
        support_candidates=texts,
        support_labels=[False for _ in texts],
        support_scores=[t[1] for t in texts_with_perplexities],
    )


async def decontext_elicit_baseline_prune_then_answer(
    paper: Paper, question: str, gold_support=None
):
    gold_support  # unused
    paper = await paper_decontext(paper)
    answer = await elicit_baseline_prune_then_answer(paper, question)
    assert answer


recipe.main(elicit_baseline_into_answer)



from fvalues import F

from ice.recipe import recipe


def make_chain_of_thought_prompt(question: str, answer_prefix: str = "") -> str:
    return F(
        f"""Answer the following question:

Question: "{question}"
Answer: "{answer_prefix}
"""
    ).strip()


async def chain_of_thought(
    question: str = "What would happen if the average temperature in Northern California went up by 5 degrees Fahrenheit?",
    answer_prefix: str = "Let's think step by step.",
) -> str:
    prompt = make_chain_of_thought_prompt(question, answer_prefix)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(chain_of_thought)



from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import list_experiments
from ice.paper import Paper
from ice.recipe import Recipe


class AllQuotesRecipe(Recipe):
    # Used to view paper parses
    # and to test whether the gold standard quotes we're looking for
    # are actually in the parses
    question_short_name_to_test = "placebo"

    async def run(self, paper: Paper):
        experiments = list_experiments(document_id=paper.document_id)

        results: list[RecipeResult] = []
        for experiment in experiments:
            gold_standard = get_gold_standard(
                document_id=paper.document_id,
                question_short_name=self.question_short_name_to_test,
                experiment=experiment,
            )

            if gold_standard:
                results.append(
                    RecipeResult(
                        experiment=experiment,
                        question_short_name=self.question_short_name_to_test,
                        document_id=paper.document_id,
                        answer="",
                        classifications=[],
                        excerpts=[
                            paragraph.__str__() for paragraph in paper.paragraphs
                        ],
                    )
                )

        self.maybe_add_to_results(results)

        return results



import logging
import os
import signal

from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.responses import FileResponse
from starlette.responses import PlainTextResponse

from ice.routes import traces
from ice.trace import traces_dir

logger = logging.getLogger(__name__)

dist_dir = Path(__file__).parent / "ui"

app = FastAPI()
app.include_router(traces.router)
app.mount(
    "/api/traces/", StaticFiles(directory=traces_dir), name="static"
)  # see comment on get_trace

try:
    app.mount("/assets/", StaticFiles(directory=dist_dir / "assets"), name="static")
except RuntimeError:
    logger.warning(
        "ui folder not found, skipping static file mount. Run `npm run build` to build the ui."
    )


PING_RESPONSE = "ought-ice says pong"


@app.get("/ping")
async def ping():
    return PlainTextResponse(PING_RESPONSE)


@app.post("/stop")
async def stop():
    # Note that this doesn't work properly when the server runs with --reload
    os.kill(os.getpid(), signal.SIGKILL)


@app.get("/{_full_path:path}")
async def catch_all(_full_path: str):
    path = dist_dir / "index.html"
    if not path.exists():
        return PlainTextResponse(
            "ui/index.html not found. Run `npm run build` in the ui directory to create it, "
            "or run `npm run dev` to run the dev server and access localhost:5173 instead."
        )
    return FileResponse(path)



from fvalues import F

from ice.recipes.primer.answer_by_dispatch.types import *


def make_action_selection_prompt(question: str) -> str:
    action_types_str = F("\n").join(
        [
            F(f"{i+1}. {action_type.description}")
            for i, action_type in enumerate(action_types)
        ]
    )

    return F(
        f"""You want to answer the question "{question}".

You have the following options:

{action_types_str}

Q: Which of these options do you want to use before you answer the question? Choose the option that will most help you give an accurate answer.
A: I want to use option #"""
    ).strip()



from fvalues import F

from ice.recipe import recipe
from ice.recipes.primer.subquestions import ask_subquestions
from ice.utils import map_async


Question = str
Answer = str
Subs = list[tuple[Question, Answer]]


def render_background(subs: Subs) -> str:
    if not subs:
        return ""
    subs_text = F("\n\n").join(F(f"Q: {q}\nA: {a}") for (q, a) in subs)
    return F(f"Here is relevant background information:\n\n{subs_text}\n\n")


def make_qa_prompt(question: str, subs: Subs) -> str:
    background_text = render_background(subs)
    return F(
        f"""{background_text}Answer the following question, using the background information above where helpful:

Question: "{question}"
Answer: "
"""
    ).strip()


async def get_subs(question: str, depth: int) -> Subs:
    subquestions = await ask_subquestions(question=question)
    subanswers = await map_async(
        subquestions, lambda q: answer_by_amplification(question=q, depth=depth)
    )
    return list(zip(subquestions, subanswers))


async def answer_by_amplification(
    question: str = "What is the effect of creatine on cognition?", depth: int = 1
):
    subs = await get_subs(question, depth - 1) if depth > 0 else []
    prompt = make_qa_prompt(question, subs=subs)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(answer_by_amplification)



from collections.abc import Callable
from collections.abc import Sequence

from ice.cache import diskcache
from ice.metrics.rouge import matches
from ice.paper import Paper
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard


@diskcache()
async def identify_gs_str(
    candidates: Sequence[str], gs_quotes: Sequence[str], lcs_threshold: float = 0.7
) -> Sequence[str]:
    positive_selections = set[str]()
    for gs_quote in gs_quotes:
        gs_matches = await matches(
            hypotheses=candidates, references=[gs_quote], lcs_threshold=lcs_threshold
        )
        for candidate in gs_matches:
            positive_selections.add(candidate)
    return [cand for cand in candidates if cand in positive_selections]


async def convert_demonstration_example(
    example: PaperQaGoldStandard,
    paper_division_func: Callable[[Paper], Sequence[str]],
) -> PaperQaGoldStandard:
    paper_parts = paper_division_func(example.paper)
    return PaperQaGoldStandard(
        paper=example.paper,
        question=example.question,
        gold_answer=example.gold_answer,
        short_gold_answer=example.short_gold_answer,
        gold_support=(await identify_gs_str(paper_parts, example.gold_support)),
    )



from collections.abc import Sequence
from itertools import chain
from typing import cast

from structlog.stdlib import get_logger

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.num_utils import extract_nums
from ice.recipes.experiments_and_arms.types import PassageWithReasoning

# TODO: Make configurable for different ways of prompting helpfulness

log = get_logger()


def make_helpfulness_prompt(prefix: str, helpful_line: str, num_excerpts: int) -> str:
    TEMPLATE = """Number of excerpts: {num_excerpts}

{helpful_line}

List the helpful excerpts here: {translation}""".strip()

    helpfulness_cases: list[dict[str, int | str | StopSentinel]] = [
        dict(
            num_excerpts=4,
            helpful_line="None of the excerpts were helpful",
            translation="None",
        ),
        dict(
            num_excerpts=4,
            helpful_line="Excerpt 3 was helpful. Excerpts 1, 2, and 4 were not helpful",
            translation="3",
        ),
        dict(
            num_excerpts=4,
            helpful_line="Excerpts 1 and 2 were helpful, and excerpt 3 was somewhat helpful. Excerpt 4 was not helpful.",
            translation="1, 2, 3",
        ),
        dict(
            num_excerpts=5,
            helpful_line="Excerpts 1, 2, and 3 were not helpful, but the rest were helpful.",
            translation="4, 5",
        ),
        dict(
            num_excerpts=3,
            helpful_line="All of the excerpts were helpful.",
            translation="1, 2, 3",
        ),
    ]
    for example in helpfulness_cases:
        example["helpful_line"] = prefix + cast(str, example["helpful_line"])
    helpfulness_cases.append(
        dict(num_excerpts=num_excerpts, helpful_line=helpful_line, translation=stop(""))
    )

    filled = format_multi(TEMPLATE, helpfulness_cases)
    return "\n\n".join(filled)


HELPFULNESS_SHARED = dict(
    thing_to_understand="how many experiments were conducted in the study"
)

HELPFULNESS_PREFIX = "Which excerpts, if any were helpful?"


async def _which_paras_were_helpful(
    helpfulness_prefix: str, helpful_line: str, num_excerpts: int
) -> list[int]:
    prompt = make_helpfulness_prompt(helpfulness_prefix, helpful_line, num_excerpts)
    completion = await recipe.agent().complete(prompt=prompt, stop="\n")
    return [num - 1 for num in extract_nums(completion)]


def extract_helpful_line(reasoning: str) -> str | None:
    lines = reasoning.split("\n")
    helpful_lines = [
        line for line in lines if "were helpful in understanding" in line.lower()
    ]
    if not helpful_lines:
        log.warning("Response missing helpful lines", reasoning=reasoning)
        return None
    if len(helpful_lines) > 1:
        log.warning(
            "Response has too many helpful lines",
            reasoning=reasoning,
            helpful_lines=helpful_lines,
        )
    return helpful_lines[0]


async def helpful_paragraphs(
    helpful_line: str, paragraphs: Sequence[str]
) -> Sequence[str]:
    idxs = await _which_paras_were_helpful(
        HELPFULNESS_PREFIX, helpful_line, len(paragraphs)
    )
    return [paragraphs[idx] for idx in idxs if idx < len(paragraphs)]


async def keep_most_helpful_paragraphs(
    passages_with_reasoning: Sequence[PassageWithReasoning],
    start_with: int = 2,
    continue_if_less_than: int = 3,
) -> Sequence[str]:
    async def get_best_paras(
        passages_with_reasoning: Sequence[PassageWithReasoning],
    ):
        helpful_lines = [pwr.helpfulness for pwr in passages_with_reasoning]
        helpful_lines_non_null = [
            (line, pwr)
            for line, pwr in zip(helpful_lines, passages_with_reasoning)
            if line
        ]
        best_paras = chain.from_iterable(
            [
                (await helpful_paragraphs(line, pwr.passage))
                for line, pwr in helpful_lines_non_null
            ]
        )
        return list(best_paras)

    initial_passages, later_passages = (
        passages_with_reasoning[:start_with],
        passages_with_reasoning[start_with:],
    )
    best_paras = set(await get_best_paras(initial_passages))
    while later_passages and len(best_paras) < continue_if_less_than:
        next_para = later_passages[0]
        later_passages = later_passages[1:]
        best_paras = best_paras | set((await get_best_paras([next_para])))

    # Sort so that the returned order is deterministic for caching etc.
    best_paras_sorted = sorted(best_paras)

    # Plain alphabetical sort actually lands on a pathological case
    # that leads to a test failure where count_experiments returns 59 which leads to
    # ValueError: Count 59 not in count word dictionary
    # So we reverse here to workaround that, although that should be solved in general
    # and the recipe/test should be moved out of the repo anyway.
    return best_paras_sorted[::-1]


recipe.main(keep_most_helpful_paragraphs)



from anyio.to_thread import run_sync
from transformers.pipelines import pipeline

from ice.agents.base import Agent


class SquadAgent(Agent):
    def __init__(self, model_name: str = "z-uo/roberta-qasper"):

        self.nlp = pipeline(
            "question-answering", model=model_name, tokenizer=model_name
        )

    async def relevance(
        self, *, question, context, verbose=False, default=None
    ) -> float:
        response = await run_sync(
            self.nlp, dict(question=question, context=context), cancellable=True
        )
        return response["score"]



from collections.abc import Mapping
from collections.abc import Sequence

from anyio.to_thread import run_sync
from pydantic import BaseModel
from rouge_metric import PyRouge

from ice.cache import diskcache
from ice.metrics.base import Metric
from ice.metrics.base import Sample


class _RougeScore(BaseModel):
    r: float
    p: float
    f: float

    def __str__(self):
        return f"recall: {self.r:.2f}, precision: {self.p:.2f}, f1: {self.f:.2f}"


class RougeResult(BaseModel):
    rouge_1: _RougeScore
    rouge_2: _RougeScore
    rouge_3: _RougeScore
    rouge_l: _RougeScore

    class Config:
        alias_generator = lambda n: n.replace("_", "-")  # noqa: E731

    def __str__(self):
        return "\n\n".join(
            [
                f"* Rouge L: {self.rouge_l}",
                f"* Rouge 1: {self.rouge_1}",
                f"* Rouge 2: {self.rouge_2}",
                f"* Rouge 3: {self.rouge_3}",
            ]
        )


class Rouge(Metric):
    name = "Rouge"

    def __init__(
        self, n: tuple[int, ...] = (1, 2, 3), rouge_l: bool = True, rouge_w: bool = True
    ):
        self.rouge = PyRouge(rouge_n=n, rouge_l=rouge_l, rouge_w=rouge_w)

    async def compute(self, sample: list[Sample]) -> list[RougeResult]:
        async def _compute_single(sample: Sample) -> RougeResult:
            hyp = sample.left
            ref = [sample.right for _ in sample.left]

            result_dict = await run_sync(
                lambda: self.rouge.evaluate(hyp, ref), cancellable=True
            )
            return RougeResult.parse_obj(result_dict)

        return [await (_compute_single(s)) for s in sample]


@diskcache()
async def rouge_texts(
    hypotheses: Sequence[str], references: Sequence[str]
) -> Mapping[str, RougeResult]:
    assert hypotheses, "hypotheses cannot be empty"
    assert references, "references cannot be empty"
    scores = await Rouge().compute(
        [Sample(left=[hyp], right=references) for hyp in hypotheses]
    )
    return {hyp: score for hyp, score in zip(hypotheses, scores)}


async def matches(
    hypotheses: Sequence[str], references: Sequence[str], lcs_threshold: float = 0.7
):
    rouge_scores = await rouge_texts(hypotheses=hypotheses, references=references)
    return {
        hyp: scores
        for hyp, scores in rouge_scores.items()
        if scores.rouge_l.r >= lcs_threshold
    }



import math

from collections.abc import Mapping
from collections.abc import Sequence
from typing import cast
from typing import TypedDict

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform

REASONINGS: list[str] = [
    """- Excerpt 1 talks about the training of medical students and trainees and so is not related to effect size.
- Excerpt 2 talks about the need for medical trainees to learn from different kinds of images to attain expert level this is also not relevant to effect size.
- Excerpt 3 talks about how individual variability substantially affects learning styles this is also not relevant to effect size.
- Excerpt 4 says that learning is dictated by the diversity and number of cases  encountered this is again relevant to effect size.
- Excerpt 5 talks the benefits of an AI assisted teaching platform. This is irrelevant to the effect size.
- Excerpt 6 talks about the WithAI score (88.87) of the medical students in the AIL group being significantly higher than the prelearning score (75.73). This directly related to the effect size.
Which excerpts answer the question, from most to least important: 6""",
    """- Excerpt 1 talks about the six experiments and states that the Ghana experiment is a clustered randomized trial. This is relevant to the design of the Ghana experiment.
- Excerpt 2 talks about the Ghana experiment having a more complex design with two additional treatment groups. This is very much relevant to the design of the Ghana experiment.
- Excerpt 3 says that in this paper, they  are using only the group that received the pooled intervention. This is not relevant to the to the design of the Ghana experiment.
- Excerpt 4 talks about the sample size used in the analysis, which could relate design of the Ghana experiment but it's not the most important detail.
- Excerpt 5 presents the baseline data for the same variables and indices used as the primary outcome measures, which is not relevant to the design of the Ghana experiment .
- Excerpt 6 says that Panel A presents the mean comparisons and t-tests for equality of means, which is not relevant to the design of the Ghana experiment.
- Excerpt 7 talks about how villages were randomly selected to be treatment or control villages in the countries with clustered randomization. This is relevant to the question.
Which excerpts answer the question, from most to least important: 2, 1, 7, 4""",
    """- Excerpt 1 talks about how information about the general health of 8-year-old singletons born through ICSI was obtained from the parents by means of a questionnaire. This is not relevant to the sample size.
- Excerpt 2 states that 15 out of 150 ICSI children experienced a major congenital malformation compared with 5/147 SC children. This excerpt includes the sample size for both the ICSI children and the SC children and is hence directly relevant to the question.
- Excerpt 3 says that pubertal staging was similar in both groups. This is not relevant to the sample size.
- Excerpt 4 says that neurological examination did not show important differences between ICSI and SC children. This is not relevant to the sample size.
- Excerpt 5 states that ICSI children did not require more remedial therapy or surgery or hospitalization than SC children. This is not relevant to the sample size.
- Excerpt 6 talks about how the medical outcome of 8-year-old singletons born through ICSI was compared with that of 147 singletons of the same age born after spontaneous conception. This excerpt includes the sample size for both the ICSI children and the SC children and relevant to the sample size.
- Excerpt 7 is the conclusion of the study and it says that physical examination including a thorough neurological examination did not reveal important differences between the two groups. This is not relevant to the sample size.
Which excerpts answer the question, from most to least important: 2, 6""",
]

EXAMPLES: list[dict[str, str | ValueTransform | StopSentinel]] = [
    dict(
        question="What was the effect size?",
        existing=numbered_list(
            [
                """This training [of medical students and trainees] requires acquiring skills to analyze and extract imaging features to identify patterns, generate a differential diagnosis that matches the patterns, and correlate the imaging features and differential with clinical findings to select the most likely diagnosis [5].""",
                """However, owing to time constraints, training opportunities might be compressed, and trainees may not be able to access as many images as their tutors or teachers [medical students and trainees need to learn several kinds of images to achieve an expert level].""",
                """Individual variability [in learning styles] is thought to substantially affect learning styles [6,7].""",
                """Moreover, learning [for medical students and trainees] is dictated by the number and diversity of cases encountered, with varying practices and patient mixes.""",
                """An artificial intelligence (AI)-assisted teaching platform can deliver personalized education and 24-h supervised tutoring that benefits both trainees [medical students and] and trainers [teachers].""",
                """The WithAI score [of the medical students in the AIL group] (88.87 ± 5.51) was significantly higher than the prelearning score [of the medical students in the AIL group] (75.73 ± 10.58, p < 0.01).""",
            ]
        ),
        reasoning_with_selection=REASONINGS[0],
        selections="6",
    ),
    dict(
        question="What was the design of the Ghana experiment?",
        existing=numbered_list(
            [
                """Of the six experiments, three are individual randomized trials with randomization at the household level within each village (India, Ethiopia and Pakistan) and three are clustered randomized trials, with randomization at both the village and household level (Ghana, Honduras, and Peru).""",
                """One site (Ghana) had a more complex design with two additional treatment groups (savings only, and productive asset grant only) to “unpack” those aspects of the intervention.""",
                """In this paper we are using only the group that received the pooled intervention.""",
                """The sample size used in the analysis varies from 925 households (Ethiopia) to 2,606 households (Ghana) from site to site.""",
                """Table S1b presents baseline data for the same variables and indices used as the primary outcome measures.""",
                """Panel A presents the mean comparisons and t-tests for equality of means.""",
                """In the countries with clustered randomization, villages were randomly selected to be treatment or control villages, and then treatment households were randomly selected within the set of eligible households in treatment villages.""",
            ]
        ),
        texts=numbered_list([]),
        reasoning_with_selection=REASONINGS[1],
        selections="2, 1, 7, 4",
    ),
    dict(
        question="What was the sample size?",
        existing=numbered_list(
            [
                """Information about their [the 8-year-old singletons born through ICSI's] general health was obtained from the parents by means of a questionnaire.""",
                """RESULTS: Fifteen of 150 ICSI children [8-year-old singletons born through ICSI] experienced a major congenital malformation compared with 5/147 SC children [of the same age born after spontaneous conception] (P < 0.05).""",
                """Pubertal staging [of 8-year-old singletons born through ICSI and of the same age born after spontaneous conception] was similar in both groups.""",
                """Neurological examination did not show important differences between ICSI [8-year-old singletons born through ICSI] and SC [of the same age born after spontaneous conception] children.""",
                """ICSI children [8-year-old singletons born through ICSI] did not require more remedial therapy or surgery or hospitalization than SC children [of the same age born after spontaneous conception].""",
                """METHODS: Medical outcome of 8-year-old singletons (n = 150) born through ICSI (³32 weeks) was compared with that of 147 singletons of the same age born after spontaneous conception (SC) [in order to investigate possible health problems in the long-term outcome of children born after ICSI].""",
                """CONCLUSION: Physical examination including a thorough neurological examination did not reveal important differences between the two groups [of 8-year-old singletons, one group born through ICSI and the other group born after spontaneous conception].""",
            ]
        ),
        reasoning_with_selection=REASONINGS[2],
        selections="6, 2",
    ),
]

INSTRUCTIONS = """Instructions: For each question and list of excerpts, record which excerpts help answer the question, from most to least important."""

EXAMPLE_TEMPLATE = """
Question: {question}

Excerpts:

{existing}

Which excerpts answer the question, from most to least important: {selections}
""".strip()

EXAMPLE_WITH_REASONING_TEMPLATE = """
Question: {question}

Excerpts:

{existing}

Let's think about how each excerpt relates to the question.

{reasoning_with_selection}
""".strip()


EXAMPLE_SEPARATOR = "\n\n---\n\n"


class RenderablePruningExample(TypedDict):
    question: str
    existing: ValueTransform[Sequence[str]]
    selections: str


def make_pruning_prompt(
    *,
    question: str,
    existing: Sequence[str],
    examples: list[RenderablePruningExample] | None = None,
) -> str:
    all_examples = (examples or EXAMPLES) + [
        dict(
            question=question,
            existing=numbered_list(existing),
            selections=stop(""),
        )
    ]
    filled_examples = format_multi(EXAMPLE_TEMPLATE, all_examples)  # type: ignore[arg-type]
    prompt = "\n\n".join((INSTRUCTIONS, EXAMPLE_SEPARATOR.join(filled_examples)))
    return prompt


def make_pruning_with_reasoning_prompt(
    *,
    question: str,
    existing: Sequence[str],
) -> str:
    all_examples = EXAMPLES[1:2] + [
        dict(
            question=question,
            existing=numbered_list(existing),
            reasoning_with_selection=stop("- Excerpt 1"),
        )
    ]
    filled_examples = format_multi(EXAMPLE_WITH_REASONING_TEMPLATE, all_examples)
    prompt = "\n\n".join((INSTRUCTIONS, EXAMPLE_SEPARATOR.join(filled_examples)))
    return prompt


class _LogProbsOpenAI(TypedDict):
    tokens: list[str]
    token_logprobs: list[float]
    top_logprobs: list[dict[str, float]]


def get_pruned_selections_via_logprobs(
    logprobs: dict, num_selections: int
) -> Mapping[int, float]:
    """Returns a mapping of scores where the
    integer part is the descending rank and
    the decimal part is the probability.

    Args:
        logprobs (dict): openai_response["choices"][0]["logprobs"]
        num_selections (int): number of candidates for pruning

    Returns:
        Mapping[int, float]: scores for each candidate, with "-inf" meaning "not present"
    """
    assert num_selections < 362
    lps = cast(_LogProbsOpenAI, logprobs)
    # Don't include tokens from reasoning
    start_idx = len(lps["tokens"]) - list(reversed(lps["tokens"])).index(" important")
    selections = [
        idx
        for idx, selection in enumerate(lps["tokens"])
        if selection.strip().isnumeric() and idx >= start_idx
    ]
    indexed_probs: dict[int, float] = {}
    for rank, selected_idx in enumerate(selections):
        for candidate in range(num_selections):
            if candidate not in indexed_probs:
                prob_at_position = lps["top_logprobs"][selected_idx].get(
                    f" {candidate + 1}"
                )
                if prob_at_position is not None:
                    indexed_probs[candidate] = (
                        len(selections) - rank + math.exp(prob_at_position)
                    )
    return {i: indexed_probs.get(i) or float("-inf") for i in range(num_selections)}


def get_pruned_selections_via_completion(completion: str) -> Sequence[int]:
    return list(map(lambda i: int(i.strip()) - 1, completion.split(",")))



from collections.abc import Sequence

from structlog.stdlib import get_logger

from ice.recipe import recipe
from ice.recipes.program_search.nodes.augment_question.prompts import EXAMPLE_SEPARATOR
from ice.recipes.program_search.nodes.augment_question.prompts import get_new_questions
from ice.recipes.program_search.nodes.augment_question.prompts import (
    make_augment_question_prompt,
)

log = get_logger()


async def augment_question(
    question: str, current_texts: Sequence[str]
) -> tuple[str, Sequence[str]]:
    """Given a question and some texts, what other questions would be helpful to answer
    to be able to answer the original question completely?

    Args:
        question (str): The original question we are trying to answer
        current_texts (Sequence[str]): Texts we currently have

    Returns:
        tuple[str, Sequence[str]]: The *most important new question, then all new questions
    """
    assert current_texts, "Augment question node not yet designed for 0-text case"
    prompt = make_augment_question_prompt(
        question=question, existing_selections=current_texts
    )
    completion = await recipe.agent().complete(
        prompt=prompt, stop=EXAMPLE_SEPARATOR, max_tokens=312
    )
    return get_new_questions(completion)


recipe.main(augment_question)



from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_full_document_id
from ice.paper import Paper
from ice.recipe import Recipe
from ice.utils import window_by_tokens

QUESTION_SHORT_NAME = "adherence"


def get_paper_text(paper: Paper) -> str:
    return "\n\n".join(str(p) for p in paper.paragraphs)


def generate_qa_prompt_instruct(paper_text: str) -> str:
    return f"""
Summarize the adherence rate of this research paper based on the following excerpt. If the adherence rate is not mentioned in this excerpt, say "not mentioned".

BEGIN PAPER EXCERPT

{paper_text}

END PAPER EXCERPT

One possible summary of the adherence rate based on the above excerpt is:""".strip()


def create_recipe_result(paper_id: str, experiment: str, answer: str) -> RecipeResult:
    return RecipeResult(
        document_id=paper_id,
        question_short_name=QUESTION_SHORT_NAME,
        experiment=experiment,
        classifications=[
            None,
            None,
        ],
        answer=answer,
        result=answer,
        excerpts=[],
    )


class FunnelSimple(Recipe):
    agent_str = "instruct"

    async def run(self, paper: Paper):
        full_paper_text = get_paper_text(paper)

        descriptions = []
        # Ask the agent to answer the prompt
        for chunk in window_by_tokens(full_paper_text, max_tokens=5000):
            description = await self.agent(self.agent_str).complete(
                prompt=generate_qa_prompt_instruct(chunk),
                max_tokens=500,
            )
            descriptions.append(description)

        # Save results for each experiment
        paper_id = get_full_document_id(paper.document_id)
        experiments: list[str] = list_experiments(document_id=paper_id)
        recipe_results = [
            create_recipe_result(paper_id, experiment, "\n---\n".join(descriptions))
            for experiment in experiments
        ]
        self.maybe_add_to_results(recipe_results)

        return "\n\n---\n\n".join(descriptions)



Name = str
Message = str
Turn = tuple[Name, Message]
Debate = list[Turn]

my_debate: Debate = [
    ("Alice", "I think we should legalize all drugs."),
    ("Bob", "I'm against."),
    ("Alice", "The war on drugs has been a failure. It's time to try something new."),
]



from dataclasses import dataclass

from ice.recipe import recipe


@dataclass
class Abstract:
    title: str
    authors: list[str]
    year: int | None
    text: str


DEFAULT_ABSTRACTS = [
    Abstract(
        title="Association Between Cigarette Smoking Prevalence and Income Level: A Systematic Review and Meta-Analysis",
        authors=[
            "Brunilda Casetta",
            "Alejandro J Videla",
            "Ariel Bardach",
            "Paola Morello",
            "Natalie Soto",
            "Kelly Lee",
            "Paul Anthony Camacho",
            "Rocío Victoria Hermoza Moquillaza",
            "Agustín Ciapponi",
        ],
        year=2016,
        text="""Introduction Previous evidence linked low socioeconomic status with higher smoking prevalence. Our objective was to assess the strength of this association in the world population, updating a previous work.
Methods Systematic review and meta-analysis of observational studies. Subgroup analyses included continents, WHO regions, country mortality levels, gender, age, risk of bias, and study publication date. Independent reviewers selected studies, assessed potential bias and extracted data. We searched MEDLINE, EMBASE, CENTRAL, SOCINDEX, AFRICAN INDEX MEDICUS, and LILACS, and other sources from 1989 to 2013 reporting direct measurements of income and current cigarette smoking.
Results We retrieved 13,583 articles and included 93 for meta-analysis. Median smoking prevalence was 17.8% (range 3-70%). Lower income was consistently associated with higher smoking prevalence (odds ratio [OR]: 1.45; 95% confidence interval [CI]: 1.35-1.56). This association was statistically significant in the subgroup analysis by WHO regions for the Americas (OR: 1.54; 95% CI: 1.42-1.68), South East Asia (OR: 1.53; 95% CI: 1.10-2.00), Europe (OR: 1.45; 95% CI: 1.29-1.63), and Western Pacific (OR: 1.32; 95% CI: 1.02-1.72), and in studies conducted during 1990s (OR: 1.42; 95% CI: 1.24-1.62) and 2000s (OR: 1.48; 95%CI: 1.30-1.64). Likewise, it was noted in low-mortality countries (OR: 1.48; 95% CI: 1.37-1.60) and for both genders. Prevalence was highest in the lowest income levels compared to the middle (OR: 1.69; 95% CI: 1.49-1.92), followed by the middle level compared to the highest (OR: 1.31; 95% CI: 1.20-1.43).
Conclusions Our results show that current cigarette smoking was significantly associated with lower income worldwide and across subgroups, suggesting a dose-response relationship.
Implications This unique updated systematic review shows a consistent inverse dose-response relationship between cigarette smoking and income level, present among most geographical areas and country characteristics. Public health measures should take into account this potential inequity and consider special efforts directed to disadvantaged populations.""",
    ),
    Abstract(
        title="Educational inequalities in cause-specific mortality in middle-aged and older men and women in eight western European populations",
        authors=[
            "Martijn Huisman",
            "Anton E Kunst",
            "Matthias Bopp",
            "Jens-Kristian Borgan",
            "Carme Borrell",
            "Giuseppe Costa",
            "Patrick Deboosere",
            "Sylvie Gadeyne",
            "Myer Glickman",
            "Chiara Marinacci",
            "Christoph Minder",
            "Enrique Regidor",
            "Tapani Valkonen",
            "Johan P Mackenbach",
        ],
        year=2005,
        text="""BACKGROUND The aim of the study was to determine whether education or income was more strongly related to smoking in the European Union at large, and within the individual countries of the EU, at the end of the 1990s.
METHODS We related smoking prevalence to education and income level by analyzing cross-sectional data on a total of 48,694 men and 52,618 women aged 16 and over from 11 countries of the European Union in 1998.
RESULTS Both education and income were related to smoking within the European Union at large. After adjustment of the other socioeconomic indicator, education remained related to smoking in the EU at large, but income only remained so among men. Educational inequalities were larger than income-related inequalities among younger and middle-aged men and women. Educational inequalities were larger than income-related inequalities among men in all individual countries, and among women in Northern Europe. For women from Southern European countries, the magnitude of education- and income-related inequalities was similar.
CONCLUSIONS Education is a strong predictor of smoking in Europe. Interventions should aim to prevent addiction to smoking among the lower educated, by price policies, school-based programs, and smoking cessation support for young adults.""",
    ),
    Abstract(
        title="Income levels and prevalence of smoking in Latin America",
        authors=[
            "Ariel Bardach",
            "Herney Andrés García Perdomo",
            "Ruth Amanda Ruano Gándara",
            "Agustín Ciapponi",
        ],
        year=2016,
        text="""Objective Determine the relationship between tobacco-use prevalence and smoker income level in Latin America and the Caribbean (LAC).
Methods A systematic search was carried out in MEDLINE, EMBASE, CENTRAL, SOCINDEX, and LILACS databases. Studies from LAC published from January 1989 to December 2015 were included and were analyzed by subgroups disaggregated by decade of data, country, bias risk, sex, and age group.
Results Of 1 254 studies evaluated by full text, 29 articles were included, of which 25 were chosen for meta-analysis. All included studies were cross-sectional or surveillance, primarily from Brazil and Mexico.Low income was associated with higher prevalence of active smoking (odds ratio [OR] 1.62; 95% confidence interval [95%CI] 1.34-1.96) than high income (reference). A dose-response effect trend was observed: middle income (OR 1.23; 95%CI 1.00-1.52) and low income (OR 1.64; 95%CI 1.17-2.30). This association was greater in men (OR 2.22; 95%CI 1.77-2.78) than in women (OR 1.6; 95%CI 1.11-2.47).
Conclusions An inverse relationship was observed between income and tobacco-use prevalence. Further efforts are required to determine this relationship in special populations, such as adolescents and pregnant women. This research can be useful for policymakers by improving tobacco control strategies and for characterizing public health equity issues.""",
    ),
    Abstract(
        title="Is income or employment a stronger predictor of smoking than education in economically less developed countries? A cross-sectional study in Hungary",
        authors=["Mall Leinsalu", "Csilla Kaposvári", "Anton E Kunst"],
        year=2011,
        text="""BackgroundIn developed European countries in the last phase of the smoking epidemic, education is a stronger predictor of smoking than income or employment. We examine whether this also applies in economically less developed countries.MethodsData from 7218 respondents in the 25-64 age group came from two National Health Interview Surveys conducted in 2000 and 2003 in Hungary. Independent effects of educational level, income and employment status were studied in relation to smoking prevalence, initiation and continuation for all age groups combined and separately for 25-34, 35-49 and 50-64 years old. Absolute levels were evaluated by using age-standardized prevalence rates. Relative differences were assessed by means of logistic regression.ResultsEducation and income, but not employment, were associated with equally large differences in smoking prevalence in Hungary in the 25-64 age group. Among men, smoking initiation was related to low educational level, whereas smoking continuation was related to low income. Among women, low education and low income were associated with both high initiation and high continuation rates. Considerable differences were found between the age groups. Inverse social gradients were generally strongest in the youngest age groups. However, smoking continuation among men had the strongest association with low income for the middle-aged group.ConclusionsPatterns of inequalities in smoking in Hungary can be best understood in relation to two processes: the smoking epidemic, and the additional effects of poverty. Equity orientated tobacco control measures should target the low educated to prevent their smoking initiation, and the poor to improve their cessation rates.""",
    ),
]


def make_prompt(abstract: Abstract, question: str) -> str:
    return f"""Answer the given question about each abstract:

###

Question 1: What is IL-13?

Title 1: Interleukin 13, an interleukin 4-like cytokine that acts on monocytes and B cells, but not on T cells

Abstract 1: Interleukin 13 (IL-13) is a recently described protein secreted by activated T cells which is a potent in vitro modulator of human monocyte and B-cell functions. The data, reviewed here by Gerard Zurawski and Jan de Vries, shows that IL-13 shares biological activities with IL-4, their genes are closely linked in both the human and mouse genomes, and there is sequence homology between IL-13 and IL-4 proteins. Although the cloned IL-4 receptor protein (IL-4R) does not bind IL-13, it appears that the functional IL-4R and IL-13R share a common subunit that is important for signal transduction.

Answer 1: Interleukin 13 (IL-13) is a recently described protein secreted by activated T cells which is a potent in vitro modulator of human monocyte and B-cell functions. It is similar to IL-4.

###

Question 2: What are the challenges in multi-agent systems?

Title 2: Multi-Agent Systems: Technical & Ethical Challenges of Functioning in a Mixed Group

Abstract 2: In today's highly interconnected, open-networked computing world, artificial intelligence computer agents increasingly interact in groups with each other and with people both virtually and in the physical world. AI's current core challenges concern determining ways to build AI systems that function effectively and safely for people and the societies in which they live. To incorporate reasoning about people, research in multi-agent systems has engendered paradigmatic shifts in computer-agent design, models, and methods, as well as the development of new representations of information about agents and their environments. These changes have raised technical as well as ethical and societal challenges. This essay describes technical advances in computer-agent representations, decision-making, reasoning, and learning methods and highlights some paramount ethical challenges.

Answer 2: The paper is not directly relevant to the question. The paper argues that research in multi-agent systems has raised societal and ethical challenges.

###

Question 3: {question}

Title 3: {abstract.title}

Abstract 3: {abstract.text}

Answer 3:"""


async def abstract_qa(
    abstract: Abstract,
    question: str,
) -> str:
    prompt = make_prompt(abstract=abstract, question=question)
    answer = await recipe.agent().complete(prompt=prompt, stop="###")
    return answer


async def abstract_qa_cli():
    abstract = DEFAULT_ABSTRACTS[0]
    question = "what is the relationship between income and smoking?"
    answer = await abstract_qa(abstract=abstract, question=question)
    return answer


recipe.main(abstract_qa_cli)



#!/usr/bin/env python
import argparse
import re

from pathlib import Path

root_dir = Path(__file__).parent.parent.parent
primer_recipe_dir = root_dir / "ice" / "ice" / "recipes" / "primer"
primer_dir = root_dir / "primer"

pattern = re.compile(
    r"""
    (?P<prefix>
        ^\{%\ code\ title="(?P<path> [^"]+)"[^%]*\ %\}\n+
        ```python\n)
    (?P<content> .*?)
    (?P<suffix> ^```\n)
    """,
    re.MULTILINE | re.DOTALL | re.VERBOSE,
)


def update_ice():
    for md_path in primer_dir.glob("**/*.md"):
        for c in pattern.finditer(md_path.read_text()):
            if " " in c["path"]:
                continue
            code_path = primer_recipe_dir / c["path"]
            code_path.parent.mkdir(parents=True, exist_ok=True)
            code_path.write_text(c["content"])


def updated_code_block(c):
    code_path = primer_recipe_dir / c["path"]
    if not code_path.exists():
        return c[0]
    return c["prefix"] + code_path.read_text() + c["suffix"]


def update_primer():
    for md_path in primer_dir.glob("**/*.md"):
        md_path.write_text(pattern.sub(updated_code_block, md_path.read_text()))


def main():
    parser = argparse.ArgumentParser(description="Update Primer code.")
    parser.add_argument(
        "--reverse",
        action="store_true",
        default=False,
        help="update ICE code based on Primer code",
    )
    args = parser.parse_args()
    if args.reverse:
        update_ice()
    else:
        update_primer()


if __name__ == "__main__":
    main()



from collections.abc import Iterable
from collections.abc import Sequence
from dataclasses import dataclass
from functools import cached_property
from itertools import chain

import matplotlib.pyplot as plt
import numpy as np

from sklearn.metrics import precision_recall_curve
from sklearn.metrics import PrecisionRecallDisplay
from sklearn.metrics import roc_auc_score

from ice.metrics.rouge import matches


@dataclass(frozen=True)
class BinaryClassificationMetrics:
    ground_truth: Sequence[bool]
    predictions: Sequence[bool]
    scores: Sequence[float] | None = None

    def __post_init__(self):
        assert len(self.ground_truth) == len(
            self.predictions
        ), "Ground truth and predictions should have the same length"

    @cached_property
    def _gt_array(self):
        return np.array(self.ground_truth, dtype=bool)

    @cached_property
    def _pred_array(self):
        return np.array(self.predictions, dtype=bool)

    @cached_property
    def tp(self):
        return (self._gt_array & self._pred_array).sum().item()

    @cached_property
    def tn(self):
        return (~self._gt_array & ~self._pred_array).sum().item()

    @cached_property
    def fn(self):
        return (self._gt_array & ~self._pred_array).sum().item()

    @cached_property
    def fp(self):
        return (~self._gt_array & self._pred_array).sum().item()

    @cached_property
    def recall(self):
        return self.tp / (self.tp + self.fn) if self.tp or self.fn else 0

    @cached_property
    def precision(self):
        return self.tp / (self.tp + self.fp) if self.tp or self.fp else 0

    @cached_property
    def f1(self):
        return (
            2 * (self.precision * self.recall) / (self.precision + self.recall)
            if self.precision or self.recall
            else 0
        )

    @cached_property
    def accuracy(self):
        return (
            (self._gt_array == self._pred_array).mean().item()
            if self.ground_truth
            else 0
        )

    @cached_property
    def auroc(self):
        if not self.scores:
            return None
        if not any(self.ground_truth):
            return None
        return roc_auc_score(y_true=self._gt_array, y_score=self.scores)

    def save_pr_curve(self, filename: str):
        if not self.scores:
            return None
        prd = PrecisionRecallDisplay.from_predictions(
            y_true=self._gt_array, y_pred=self.scores
        )
        prd.plot()
        plt.savefig(filename)

    def pr_thresholds(self, n: int | None = 20):
        if not self.scores:
            return None
        precisions, recalls, thresholds = precision_recall_curve(
            y_true=self._gt_array, probas_pred=self.scores
        )
        n = min(n or len(thresholds), len(thresholds))
        idxs = np.linspace(0, len(thresholds), n, endpoint=False).astype(int)
        return {
            float(t): dict(p=float(p), r=float(r))
            for p, r, t in zip(precisions[idxs], recalls[idxs], thresholds[idxs])
        }

    def as_dict(self) -> dict[str, int | float | None]:
        return dict(
            tp=int(self.tp),
            tn=int(self.tn),
            fp=int(self.fp),
            fn=int(self.fn),
            recall=float(self.recall),
            precision=float(self.precision),
            f1=float(self.f1),
            accuracy=float(self.accuracy),
            auroc=self.auroc and float(self.auroc),
        )

    def __str__(self):
        return str(self.as_dict())

    def __repr__(self):
        return f"<BinaryClassificiationMetrics {str(self)}>"

    @classmethod
    def aggregate(
        cls, metrics: Iterable["BinaryClassificationMetrics"]
    ) -> "BinaryClassificationMetrics":
        ms = list(metrics)
        do_scores = all((m.scores is not None for m in ms))
        return cls(
            ground_truth=list(chain(*(m.ground_truth for m in ms))),
            predictions=list(chain(*(m.predictions for m in ms))),
            scores=list(chain(*(m.scores or [] for m in ms))) if do_scores else None,
        )


async def label_texts(
    texts: Sequence[str], golds: Sequence[str], lcs_threshold: float = 0.7
) -> Sequence[bool]:
    gs_matches: set[str] = set()
    for gold in golds:
        gs_matches = gs_matches.union(
            (
                await matches(
                    hypotheses=texts, references=[gold], lcs_threshold=lcs_threshold
                )
            ).keys()
        )
    return [text in gs_matches for text in texts]


async def fuzzy_text_classification_metrics(
    texts: Sequence[str],
    predictions: Sequence[bool],
    ground_truth: Sequence[str],
    scores: Sequence[float] | None = None,
    lcs_threshold: float = 0.7,
) -> BinaryClassificationMetrics:
    """Because labeled ground truths are often partial excerpts, use Rouge lcs-recall of ground truth to generate labels.

    Args:
        texts (Sequence[str]): The texts in question
        predictions (Sequence[bool]): Labels to evaluate
        ground_truth (Sequence[str]): The positive strings to be matched with texts using rouge-lcs-recall

    Returns:
        BinaryClassificationMetrics: Metrics
    """
    gt_labels = await label_texts(texts, ground_truth, lcs_threshold=lcs_threshold)
    return BinaryClassificationMetrics(
        ground_truth=gt_labels, predictions=predictions, scores=scores
    )



import ast
import asyncio

from sys import argv

import pandas as pd

from ice.environment import env
from ice.evaluation.evaluate_recipe_result import EvaluatedRecipeResult
from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.evaluation.evaluation_report import EvaluationReport
from ice.evaluation.utils import CSVS_PATH
from ice.evaluation.utils import start_time
from ice.utils import map_async


async def summarize_experiment_evals(results_file: str):
    data_df = pd.read_csv(results_file)
    # data_df["recipe"] = "In-app QA"

    eval_dfs = [
        df for _, df in data_df.groupby(["recipe", "elicit_commit"], dropna=False)
    ]
    dashboard_row_dfs = []
    experiment_evaluations_dfs = []

    for recipe_df in eval_dfs:
        recipe_results_for_evaluation = [
            RecipeResult(
                question_short_name=row.question_short_name,
                document_id=row.document_id,
                answer="" if pd.isna(row.answer) else row.answer,
                experiment=row.experiment,
                excerpts=ast.literal_eval(row.excerpts),
                classifications=[
                    row.get("classification_1"),
                    row.get("classification_2"),
                ],
                answer_rating=None
                if pd.isna(row.get("answer_rating"))
                else int(row.get("answer_rating")),
                elicit_commit=row.get("elicit_commit"),
                failure_modes=None
                if pd.isna(row.get("failure_modes"))
                else row.failure_modes.split(","),
            )
            for _, row in recipe_df.iterrows()
        ]

        evaluation_report = EvaluationReport(
            technique_name=recipe_df["recipe"].iloc[0],
            results=await map_async(
                recipe_results_for_evaluation, EvaluatedRecipeResult.from_recipe_result
            ),
        )

        env().print(
            str(evaluation_report), format_markdown=True, wait_for_confirmation=True
        )

        experiment_evaluations_dfs.append(
            evaluation_report.make_experiments_evaluation_df()
        )
        dashboard_row_dfs.append(evaluation_report.make_dashboard_row_df())

    if len(eval_dfs) > 1:
        dashboard_rows_df = pd.concat(dashboard_row_dfs)
        recipes_str = " ".join(data_df.recipe.unique())
        dashboard_rows_df.to_csv(
            CSVS_PATH / f"dashboard_rows {recipes_str} {start_time}.csv"
        )

        experiment_evaluations_df = pd.concat(experiment_evaluations_dfs)
        experiment_evaluations_df.to_csv(
            CSVS_PATH / f"experiment_evaluations {recipes_str} {start_time}.csv"
        )


if __name__ == "__main__":
    asyncio.run(summarize_experiment_evals(argv[1]))



import math

from typing import Any

from structlog.stdlib import get_logger

from ice.agents.base import Agent
from ice.agents.base import Stop
from ice.apis.openai import openai_complete
from ice.environment import env
from ice.utils import longest_common_prefix

log = get_logger()


class OpenAIAgent(Agent):
    """An agent that uses the OpenAI API to generate answers and predictions."""

    def __init__(
        self,
        model: str = "text-davinci-002",
        temperature: float = 0.0,
        top_p: float = 1.0,
    ):
        self.model = model
        self.temperature = temperature
        self.top_p = top_p

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ) -> str:
        """Generate an answer to a question given some context."""
        if verbose:
            self._print_markdown(prompt)
        response = await self._complete(prompt, stop=stop, max_tokens=max_tokens)
        completion = self._extract_completion(response)
        if verbose:
            self._print_markdown(completion)
        return completion

    async def predict(self, *, context, default="", verbose=False) -> dict[str, float]:
        """Generate a probability distribution over the next token given some context."""
        if verbose:
            self._print_markdown(context)
        response = await self._complete(context, logprobs=5, max_tokens=1)
        prediction = self._extract_prediction(response)
        if verbose:
            self._print_markdown(prediction)
        return prediction

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: str | None = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], str | None]:
        """Generate a classification from a list of choices given some context and a question."""
        if verbose:
            self._print_markdown(prompt)
            self._print_markdown(choices)

        choice_prefix = longest_common_prefix(choices).rstrip()
        prompt_with_prefix = f"{prompt}{choice_prefix}"

        if prompt_with_prefix.endswith(" "):
            prompt_with_prefix = prompt_with_prefix[:-1]
            default = " "
        else:
            default = ""

        prediction = await self.predict(context=prompt_with_prefix, default=default)

        rel_probs = self._compute_relative_probs(choices, choice_prefix, prediction)

        if verbose:
            self._print_markdown(rel_probs)

        return rel_probs, None

    async def _complete(self, prompt, **kwargs) -> dict:
        """Send a completion request to the OpenAI API with the given prompt and parameters."""
        kwargs.update(
            {
                "model": self.model,
                "temperature": self.temperature,
                "top_p": self.top_p,
                "n": 1,
            }
        )
        response = await openai_complete(prompt, **kwargs)
        if "choices" not in response:
            raise ValueError(f"No choices in response: {response}")
        return response

    def _extract_completion(self, response: dict) -> str:
        """Extract the answer text from the completion response."""
        return response["choices"][0]["text"].strip()

    def _extract_prediction(self, response: dict) -> dict[str, float]:
        """Extract the prediction dictionary from the completion response."""
        answer = response["choices"][0]["logprobs"]["top_logprobs"][0]
        return {k: math.exp(p) for (k, p) in answer.items()}

    def _compute_relative_probs(
        self, choices: tuple[str, ...], choice_prefix: str, prediction: dict[str, float]
    ) -> dict[str, float]:
        """Compute the relative probabilities of the choices based on the prediction."""

        def lookup_prob(choice: str):
            scores = 0.0
            for (token, prob) in prediction.items():
                if choice[len(choice_prefix) :].startswith(token):
                    scores += prob
            return scores

        abs_probs = {choice: lookup_prob(choice) for choice in choices}
        Z = sum(abs_probs.values())
        if Z < 0.6:
            log.warning(f"{1-Z} of unaccounted probability in classify")
            log.warning(choice_prefix)
            log.warning(str(prediction))
            log.warning(str(abs_probs))

        rel_probs = (
            {choice: prob / Z for (choice, prob) in abs_probs.items()}
            if Z != 0.0
            else abs_probs
        )
        return rel_probs

    def _print_markdown(self, obj: Any):
        """Print the text with markdown formatting."""
        env().print(obj, format_markdown=True)



import re

from collections.abc import Sequence
from typing import Optional

from structlog import get_logger

from ice.apis.openai import openai_complete
from ice.recipes.program_search.nodes.answer.types import Demonstration


log = get_logger()


_NA_PHRASE = "not mentioned in the excerpt"
COMBINED_NA_PHRASE = f"The answer to the question is {_NA_PHRASE}"

# Instruct sometimes says "mentioned in the excerpt"
# That's not a useful output, so we want to consider it a non-answer
_NA_MATCH_PHRASE = "mentioned in the excerpt"


async def answer_like_elicit_qa(
    *,
    question: str,
    passage: str,
) -> str:

    prompt = elicit_qa_prompt(
        qa_question=question,
        excerpt=passage,
    )

    response = await openai_complete(prompt, stop=None)

    choices = response.get("choices")

    response_text = choices[0]["text"].strip()
    answer = _process_instruct_answer(response_text)

    if answer is None:
        return "Not mentioned"

    return answer


def convert_to_non_answer(example: Demonstration) -> Demonstration:
    return Demonstration(
        question=example.question, texts=example.texts, answer=COMBINED_NA_PHRASE
    )


def make_few_shot_examples(
    examples: Sequence[Demonstration],
) -> Sequence[tuple[str, str]]:
    return [
        (
            elicit_qa_prompt(
                qa_question=e.question, excerpt="\n\n".join((t for t in e.texts))
            ),
            "".join((" ", e.answer.strip())),
        )
        for e in examples
    ]


def elicit_qa_prompt(
    *,
    qa_question: str,
    excerpt: str,
) -> str:

    full_answer_prefix = "Answer:"

    return f"""Answer the question "{qa_question}" based on the excerpt from a research paper. \
Try to answer, but say "{COMBINED_NA_PHRASE}" if you don't know how to answer. \
Include everything that the paper excerpt has to say about the answer. \
Make sure everything you say is supported by the excerpt. \
The excerpt may cite other papers; \
answer about the paper you're reading the excerpt from, not the papers that it cites. \
Answer in one phrase or sentence:

Paper excerpt: {excerpt}

Question: {qa_question}

{full_answer_prefix}"""


def _process_instruct_answer(text: str) -> Optional[str]:
    text = text.strip()
    if re.search(_NA_MATCH_PHRASE, text, re.IGNORECASE):
        return None

    text = re.sub(r"^the |^that |^is |\.$", "", text)

    return text



from collections.abc import Callable
from collections.abc import Sequence
from statistics import mean

from pydantic import BaseModel

from ice.evaluation.utils import rouge_compare
from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import GoldStandard
from ice.settings import settings
from ice.utils import map_async


class MatchResult(BaseModel):
    text: str
    found: bool


class EvaluatedExcerpts(BaseModel):
    gold_standards_in_excerpts_results: Sequence[MatchResult]
    excerpts: Sequence[str]
    average_recall: float | None

    @classmethod
    async def from_excerpts_and_gold_quotes(
        cls,
        excerpts: Sequence[str],
        gold_quotes: Sequence[str],
    ) -> "EvaluatedExcerpts":
        all_excerpts = " |||| ".join(excerpts)

        gold_standards_scores = await map_async(
            gold_quotes,
            lambda gold_standard: rouge_compare([all_excerpts], [gold_standard]),
        )

        gold_standards_recalls = [score.rouge_l.r for score in gold_standards_scores]

        gold_standards_found = [
            recall > settings.GS_QUOTE_FOUND_THRESHOLD
            for recall in gold_standards_recalls
        ]

        overall_recall = (
            mean(gold_standards_recalls) if gold_standards_recalls else None
        )

        return cls(
            gold_standards_in_excerpts_results=[
                MatchResult(text=gold_standard, found=found)
                for gold_standard, found in zip(gold_quotes, gold_standards_found)
            ],
            excerpts=excerpts,
            average_recall=overall_recall,
        )

    def summary_stats_str(self) -> str:
        if not self.excerpts:
            return "The recipe didn't find any excerpts. This usually means that the recipe didn't try to find excerpts and instead answered by reading the whole paper."

        found_count = len(
            [
                gold_standard
                for gold_standard in self.gold_standards_in_excerpts_results
                if gold_standard.found
            ]
        )
        total_count = len(self.gold_standards_in_excerpts_results)
        proportion = self.proportion_gold_standards_found
        return f"""The recipe found {found_count}/{total_count}{' (' + f'{proportion:.0%}' + ')' if proportion is not None else ''} gold standard excerpts."""

    def gold_standards_str(self) -> str:
        return "\n".join(
            f"{i+1}. {'(Found by recipe)' if gold_standard_match.found else '(Not found by recipe)'} {gold_standard_match.text}"
            for i, gold_standard_match in enumerate(
                self.gold_standards_in_excerpts_results
            )
        )

    def excerpts_str(self) -> str:
        if self.excerpts:
            return "\n\nRecipe excerpts:\n\n" + "\n".join(
                f"{i}. {excerpt}" for i, excerpt in enumerate(self.excerpts, 1)
            )
        return ""

    def __str__(self) -> str:
        return f"""{self.summary_stats_str()}

Gold standard excerpts:

{self.gold_standards_str()}{self.excerpts_str()}

Average ROUGE-L recall across all gold standards (for debugging): {self.average_recall}
"""

    @property
    def num_gold_standards_found(self) -> int | None:
        if len(self.gold_standards_in_excerpts_results) == 0:
            return None
        return len(
            [
                gold_standard
                for gold_standard in self.gold_standards_in_excerpts_results
                if gold_standard.found
            ]
        )

    @property
    def proportion_gold_standards_found(self) -> float | None:
        num_gold_standards_found = self.num_gold_standards_found
        if (
            num_gold_standards_found is None
            or len(self.gold_standards_in_excerpts_results) == 0
        ):
            return None
        return num_gold_standards_found / len(self.gold_standards_in_excerpts_results)


class EvaluatedClassification(BaseModel):
    predicted: str | None
    gold: str | None
    classification_eq: Callable[[str | None, str | None], bool | None] | None

    @property
    def is_correct(self) -> bool | None:
        if self.classification_eq is None:
            if self.gold is None:
                return None
            return self.predicted == self.gold
        else:
            return self.classification_eq(self.predicted, self.gold)

    def __str__(self) -> str:
        correctness = (
            "Correct"
            if self.is_correct == True
            else "Incorrect"
            if self.is_correct == False
            else "Not evaluated"
        )
        return f"""{correctness}.
    - Predicted: {self.predicted}
    - Gold: {self.gold}"""


class RecipeResult(BaseModel):
    question_short_name: str
    document_id: str
    answer: str
    experiment: str
    excerpts: Sequence[str]
    result: object | None = None
    classifications: Sequence[str | None] = []
    classification_eq: Sequence[
        Callable[[str | None, str | None], bool | None] | None
    ] = []
    elicit_commit: str | None
    answer_rating: int | None
    failure_modes: Sequence[str] | None


class EvaluatedRecipeResult(RecipeResult):
    evaluated_excerpts: EvaluatedExcerpts
    gold_standard: GoldStandard | None

    @classmethod
    async def from_recipe_result(
        cls, recipe_result: RecipeResult
    ) -> "EvaluatedRecipeResult":
        gold_standard = get_gold_standard(
            document_id=recipe_result.document_id,
            question_short_name=recipe_result.question_short_name,
            experiment=recipe_result.experiment,
        )
        return cls(
            gold_standard=gold_standard,
            evaluated_excerpts=await EvaluatedExcerpts.from_excerpts_and_gold_quotes(
                excerpts=recipe_result.excerpts,
                gold_quotes=gold_standard.quotes if gold_standard else [],
            ),
            **recipe_result.dict(),
        )

    @property
    def evaluated_classifications(self) -> list[EvaluatedClassification]:
        recipe_classifications = self.classifications
        gold_classifications = (
            self.gold_standard.classifications if self.gold_standard else []
        )

        evaluated_classifications = []

        for i in range(0, max(len(recipe_classifications), len(gold_classifications))):
            evaluated_classification = EvaluatedClassification(
                predicted=recipe_classifications[i]
                if i < len(recipe_classifications)
                else None,
                gold=gold_classifications[i] if i < len(gold_classifications) else None,
                classification_eq=self.classification_eq[i]
                if i < len(self.classification_eq)
                else None,
            )

            evaluated_classifications.append(evaluated_classification)

        return evaluated_classifications

    def answer_str(self) -> str:
        return f"""- Answer: {self.answer}
- Gold standard: {self.gold_standard.answer if self.gold_standard else 'n/a'}
"""

    def classifications_str(self) -> str:
        return "\n".join(
            [
                f"- Classification {i+1}: {evaluated_classification}"
                for i, evaluated_classification in enumerate(
                    self.evaluated_classifications
                )
            ]
        )

    def __str__(self) -> str:
        return f"""
## {self.document_id} / {self.question_short_name} / {self.experiment}

### Answer

{self.answer_str()}

### Classifications

{self.classifications_str()}

### Finding excerpts

{self.evaluated_excerpts}
"""



from types import SimpleNamespace
from typing import Any

from pydantic import BaseModel
from structlog.stdlib import get_logger

from ice.agent import Agent
from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import get_gold_standards
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_full_document_id
from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import Recipe
from ice.utils import flatten
from ice.utils import longest_common_prefix
from ice.utils import map_async

log = get_logger()

Experiment = str


def get_section_title(paragraph: Paragraph) -> str:
    if not paragraph.sections:
        return "Unknown section"
    return paragraph.sections[0].title


def parse_quotes(s: str) -> list[str]:
    if s.strip() == "n/a":
        return []
    return [clean_quote for line in s.split("\n") if (clean_quote := line.strip('" '))]


def make_placebo_kind_aggregation_prompt(
    experiment: str, quotes: list[str], descriptions: list[str]
) -> str:
    quotes_str = "\n".join(f'- "{quote}"' for quote in quotes)
    description_str = "\n".join(f"- {description}" for description in descriptions)
    prompt = f"""Rewrite a description of the placebo used in the "{experiment}" experiment of a scientific paper.

Quotes from the paper that mention the placebo:
{quotes_str}

Descriptions of the placebo used in the "{experiment}" experiment:
{description_str}

Rewritten description of the placebo used in the "{experiment}" experiment (without repetition, using only information from the quotes and initial description): The placebo was"""
    return prompt


class DialogState(BaseModel):
    agent: Agent
    context: str
    verbose: bool = False

    async def ask(self, question: str, multiline=True, answer_prefix=""):
        partial_answer = await self.agent.complete(
            prompt=f"{self.context}\n\nQ: {question}\n\nA: {answer_prefix}",
            stop=None if multiline else "\n",
        )
        answer = f"{answer_prefix} {partial_answer}".strip()
        successor_context = f"{self.context}\n\nQ: {question}\n\nA: {answer}"
        new_state = DialogState(
            agent=self.agent, context=successor_context, verbose=self.verbose
        )
        if self.verbose:
            new_state.pprint()
        return answer, new_state

    async def multiple_choice(
        self, question: str, answers: list[str]
    ) -> tuple[dict[str, float], "DialogState"]:

        answer_prefix = longest_common_prefix(answers).rstrip()
        new_context = f"{self.context}\n\nQ: {question}\n\nA: {answer_prefix}"
        prediction = await self.agent.predict(context=new_context, default=" ")

        def lookup_prob(answer: str):
            scores = 0.0
            for (token, prob) in prediction.items():
                if answer[len(answer_prefix) :].startswith(token):
                    scores += prob
            return scores

        abs_probs = {answer: lookup_prob(answer) for answer in answers}
        Z = sum(abs_probs.values())
        if Z < 0.8:
            log.warning(f"{1-Z} of unaccounted probability in multiple choice")
            log.warning(answer_prefix)
            log.warning(str(prediction))
            log.warning(str(abs_probs))
        rel_probs = {answer: prob / Z for (answer, prob) in abs_probs.items()}
        most_likely_answer = max(answers, key=lambda answer: rel_probs[answer])

        successor_context = new_context[: -len(answer_prefix)] + most_likely_answer
        new_state = DialogState(
            agent=self.agent, context=successor_context, verbose=self.verbose
        )
        if self.verbose:
            new_state.pprint()

        return rel_probs, new_state

    def pprint(self):
        # rich.print(Panel(self.context))
        s = f"""
{'-'*80}
{self.context}
{'-'*80}
"""
        # env().print(s, format_markdown=True)
        print(s)

    class Config:
        arbitrary_types_allowed = True


def make_initial_paragraph_context(
    paragraph: Paragraph, experiment: str, section_title: str
) -> str:
    prompt = f"""Below is a paragraph from the "{section_title}" section of a scientific paper. We'll answer questions about potential use of a placebo in the "{experiment}" experiment.

The paragraph: "{paragraph}"
""".strip()
    return prompt


class PlaceboDialogs(Recipe):

    verbose = False

    msg = SimpleNamespace(
        **dict(
            relevant="The paragraph mentions the placebo procedure.",
            irrelevant="The paragraph doesn't mention the placebo procedure.",
            open_control="The paragraph states that the experiment was open-control/open-label.",
            no_open_control="The paragraph doesn't state that the experiment was open-control/open-label.",
            explicit_placebo="The paragraph states that the experiment used a placebo.",
            not_explicit_placebo="The paragraph doesn't state that the experiment used a placebo.",
            explicit_no_placebo="The paragraph states that the experiment didn't use a placebo.",
            not_explicit_no_placebo="The paragraph doesn't state that the experiment didn't use a placebo.",
            placebo_described="The kind of placebo is described.",
            placebo_undescribed="The kind of placebo is not described.",
            explicit_control="The paragraph states that there was a control group.",
            not_explicit_control="The paragraph doesn't state that there was a control group.",
            explicit_control_treatment="The paragraph says whether the control group was treated or not.",
            not_explicit_control_treatment="The paragraph doesn't say whether the control group was treated or not.",
            treatment_nothing="The control group treatment was nothing.",
            treatment_placebo="The control group treatment was a placebo.",
            treatment_baseline="The control group treatment was another baseline treatment.",
            treatment_unspecified="The control group treatment was unspecified.",
        )
    )

    async def analyze_paragraph(self, paragraph: Paragraph, experiment: Experiment):
        section_title = get_section_title(paragraph)
        initial_paragraph_context = make_initial_paragraph_context(
            paragraph, experiment, section_title
        )

        msg = self.msg

        state = DialogState(
            agent=self.agent(), context=initial_paragraph_context, verbose=self.verbose
        )
        return_info: dict[str, Any] = {"section": section_title}

        # Ask about the control group

        control_placebo_prob: float | None = None
        control_no_placebo_prob: float | None = None
        explicit_control_quotes = []

        explicit_control_p, state = await state.multiple_choice(
            question=f"""Does the paragraph state that there was a control group in the "{experiment}" experiment? If yes, say "{msg.explicit_control}". If no, say "{msg.not_explicit_control}".""",
            answers=[msg.explicit_control, msg.not_explicit_control],
        )

        if explicit_control_p[msg.explicit_control] > 0.5:
            explicit_control_quotes_str, state = await state.ask(
                question=f"""What quotes from the paragraph mention the control group in the "{experiment}" experiment? Say n/a if none are relevant to the control group""",
            )

            control_treatment_p, state = await state.multiple_choice(
                question=f"""Does the paragraph say whether the control group was treated or not? If yes, say "{msg.explicit_control_treatment}". If no, say "{msg.not_explicit_control_treatment}".""",
                answers=[
                    msg.explicit_control_treatment,
                    msg.not_explicit_control_treatment,
                ],
            )

            if control_treatment_p[msg.explicit_control_treatment] > 0.5:
                control_treatment_p, state = await state.multiple_choice(
                    question=f"""Which of the following is true: The control group treatment was {msg.treatment_nothing} / {msg.treatment_placebo} / {msg.treatment_baseline} / {msg.treatment_unspecified}?""",
                    answers=[
                        msg.treatment_nothing,
                        msg.treatment_placebo,
                        msg.treatment_baseline,
                        msg.treatment_unspecified,
                    ],
                )

                control_placebo_prob = control_treatment_p[msg.treatment_placebo]

                # "nothing" and "another baseline" imply no placebo
                control_no_placebo_prob = max(
                    control_treatment_p[msg.treatment_nothing],
                    control_treatment_p[msg.treatment_baseline],
                )

                if control_placebo_prob > 0.5 or control_no_placebo_prob > 0.5:
                    explicit_control_quotes = parse_quotes(explicit_control_quotes_str)

        # Ask whether control was open

        open_control_p, state = await state.multiple_choice(
            question=f"""Does the paragraph state that the "{experiment}" experiment was "open-control" or "open-label"? If yes, say "{msg.open_control}". If no, say "{msg.no_open_control}".""",
            answers=[msg.open_control, msg.no_open_control],
        )

        if open_control_p[msg.open_control] < 0.5:
            open_control_quotes = []
        else:
            open_control_quotes_str, state = await state.ask(
                "Which quotes from the paragraph explicitly talk about whether the study was open-control/open-label? One per line. Say n/a if there are none."
            )
            open_control_quotes = parse_quotes(open_control_quotes_str)

        # Ask about quotes that talk about whether a placebo was used

        used_placebo_quotes_str, state = await state.ask(
            "Which quotes from the paragraph talk about whether a placebo was used or not? One per line. Say n/a if none are relevant to placebos."
        )
        used_placebo_quotes = parse_quotes(used_placebo_quotes_str)

        explicit_placebo_p, state = await state.multiple_choice(
            question=f"""Does the paragraph state explicitly that the "{experiment}" experiment used a placebo? If yes, say "{msg.explicit_placebo}". If no, say "{msg.not_explicit_placebo}".""",
            answers=[msg.explicit_placebo, msg.not_explicit_placebo],
        )

        explicit_no_placebo_p, state = await state.multiple_choice(
            question=f"""Does the paragraph state explicitly that the "{experiment}" experiment didn't use a placebo? If yes, say "{msg.explicit_no_placebo}". If no, say "{msg.not_explicit_no_placebo}".""",
            answers=[msg.explicit_no_placebo, msg.not_explicit_no_placebo],
        )

        no_placebo_prob = max(
            open_control_p[msg.open_control],
            explicit_no_placebo_p[msg.explicit_no_placebo],
            control_no_placebo_prob or 0.0,
        )
        placebo_prob = max(
            explicit_placebo_p[msg.explicit_placebo],
            control_placebo_prob or 0.0,
        )

        return_info["used_placebo"] = {
            "quotes": list(
                set(used_placebo_quotes + open_control_quotes + explicit_control_quotes)
            ),
            "components": {
                "open_control": open_control_p[msg.open_control],
                "explicit_placebo": explicit_placebo_p[msg.explicit_placebo],
                "explicit_no_placebo": explicit_no_placebo_p[msg.explicit_no_placebo],
                "control_placebo": control_placebo_prob,
                "control_no_placebo": control_no_placebo_prob,
            },
            "placebo_prob": placebo_prob,
            "no_placebo_prob": no_placebo_prob,
        }

        if placebo_prob < 0.5:
            return return_info

        # Ask about quotes that talk about what the placebo was

        placebo_kind_quotes_str, state = await state.ask(
            f"""Which quotes from the paragraph describe what the placebo was in the "{experiment}" experiment? One per line. Say n/a if none describe the placebo."""
        )
        placebo_kind_quotes = parse_quotes(placebo_kind_quotes_str)

        # Ask about whether the quotes describe what the placebo was

        placebo_kind_p, state = await state.multiple_choice(
            f"""Does the paragraph say what kind of placebo was used in the "{experiment}" experiment? Say "{msg.placebo_described}" if it explains something about the placebo. Say "{msg.placebo_undescribed}" if it doesn't say what the placebo was like.""",
            answers=[msg.placebo_described, msg.placebo_undescribed],
        )

        return_info["placebo_kind"] = {
            "quotes": placebo_kind_quotes,
            "probs": placebo_kind_p,
        }
        if placebo_kind_p[msg.placebo_described] < 0.5:
            return return_info

        # Ask what the placebo was

        placebo_kind, state = await state.ask(
            f"""What does the paragraph say about what the placebo was like in the "{experiment}" experiment?"""
        )

        return_info["placebo_kind"]["answer"] = placebo_kind
        return return_info

    @staticmethod
    async def aggregate_used_placebo(paragraph_infos: list[dict]) -> dict:
        aggregate = {}

        pro_placebo_infos = []
        con_placebo_infos = []
        for paragraph_info in paragraph_infos:
            if "used_placebo" in paragraph_info:
                used_placebo_prob = paragraph_info["used_placebo"]["placebo_prob"]
                no_placebo_prob = paragraph_info["used_placebo"]["no_placebo_prob"]
                if used_placebo_prob > 0.5:
                    pro_placebo_infos.append(paragraph_info["used_placebo"])
                if no_placebo_prob > 0.5:
                    con_placebo_infos.append(paragraph_info["used_placebo"])
        if pro_placebo_infos and not con_placebo_infos:
            aggregate = {
                "answer": "Placebo",
                "quotes": flatten([info["quotes"] for info in pro_placebo_infos]),
            }
        elif not pro_placebo_infos and con_placebo_infos:
            aggregate = {
                "answer": "No placebo",
                "quotes": flatten([info["quotes"] for info in con_placebo_infos]),
            }
        elif pro_placebo_infos and con_placebo_infos:
            aggregate = {
                "answer": "Unclear",
                "quotes": flatten(
                    [info["quotes"] for info in con_placebo_infos + pro_placebo_infos]
                ),
                "quotes_pro": [info["quotes"] for info in pro_placebo_infos],
                "quotes_cons": [info["quotes"] for info in con_placebo_infos],
            }
        else:
            aggregate = {"answer": "Not mentioned", "quotes": []}

        return aggregate

    async def aggregate_placebo_kind(
        self, paragraph_infos: list[dict], experiment: str
    ) -> dict:
        answers = []
        quotes = []

        for paragraph_info in paragraph_infos:
            if "placebo_kind" in paragraph_info:
                kind = paragraph_info["placebo_kind"]
                if "answer" in kind:
                    answers.append(kind["answer"])
                    quotes += kind["quotes"]
        prompt = make_placebo_kind_aggregation_prompt(experiment, quotes, answers)

        answer_completion = await self.agent().complete(
            prompt=prompt,
        )
        answer = f"The placebo was {answer_completion}"

        if self.verbose:
            print(f"{prompt} {answer_completion}")

        return {"answer": answer, "quotes": quotes, "component_answers": answers}

    async def analyze_experiment(self, paper: Paper, experiment: Experiment):

        paragraphs = [
            paragraph
            for paragraph in paper.paragraphs
            if str(paragraph).strip()
            and len(str(paragraph)) > 40  # Against hallucination
        ]

        # Run paragraph-wise dialogs
        paragraph_infos = await map_async(
            paragraphs,
            lambda paragraph: self.analyze_paragraph(paragraph, experiment),
            max_concurrency=10 if self.mode == "machine" else 1,
            show_progress_bar=True,
        )

        # Aggregate paragraph information
        aggregate_used = await self.aggregate_used_placebo(paragraph_infos)
        aggregate_kind = await self.aggregate_placebo_kind(paragraph_infos, experiment)

        # Summarize aggregates to match RecipeResult schema
        has_placebo_info = aggregate_used["answer"] == "Placebo"
        placebo_result = (
            aggregate_kind["answer"] if has_placebo_info else aggregate_used["answer"]
        )
        placebo_excerpts = list(
            set(aggregate_kind["quotes"] + aggregate_used["quotes"])
        )

        # Log for analysis
        self.maybe_add_to_results(
            [
                RecipeResult(
                    document_id=paper.document_id,
                    question_short_name="placebo",
                    experiment=experiment,
                    classifications=[
                        aggregate_used["answer"],
                        "Placebo"
                        if has_placebo_info
                        else "No placebo or placebo not mentioned",
                    ],
                    answer=placebo_result,
                    result=placebo_result,
                    excerpts=placebo_excerpts,
                )
            ]
        )

        gold_standards = get_gold_standards(
            document_id=paper.document_id,
            question_short_name="placebo",
            experiment=experiment,
        )

        if gold_standards:
            gold_standard = gold_standards[0]
            gold_answer = gold_standard.answer
            gold_quotes = gold_standard.quotes
        else:
            gold_answer = "n/a"
            gold_quotes = []

        return {
            "result": {"actual": placebo_result, "gold": gold_answer},
            "quotes": {"actual": placebo_excerpts, "gold": gold_quotes},
            "paragraph_infos": paragraph_infos,
            "aggregate": {
                "used_placebo": aggregate_used,
                "placebo_kind": aggregate_kind,
            },
        }

    async def run(self, paper: Paper):
        experiments = list_experiments(
            document_id=get_full_document_id(paper.document_id)
        )

        results_by_experiment = {}
        for experiment in experiments:
            experiment_result = await self.analyze_experiment(paper, experiment)
            results_by_experiment[experiment] = experiment_result

        return results_by_experiment



from typing import Any
from typing import Literal

from pydantic import BaseModel
from rich.pretty import pprint
from structlog.stdlib import get_logger

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_full_document_id
from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import Recipe
from ice.recipes.comparisons_qa import ComparisonsQA
from ice.recipes.experiment_arms import ExperimentArms
from ice.utils import map_async
from ice.utils import max_by_value

Experiment = str

log = get_logger()

DEFINITIONS_STRING = """
- Placebo-controlled study: A way of testing a treatment in which a separate control group receives a sham "placebo" treatment which is specifically designed to be indistinguishable from the real treatment but has no real effect.
- Open-label trial, or open trial: A type of clinical trial in which information is not withheld from trial participants. In particular, both the researchers and participants know which treatment is being administered.
- Double-blind study: A type of clinical trial in which neither the participants nor the researcher knows which treatment or intervention participants are receiving until the clinical trial is over.
""".strip()


class Component(BaseModel):
    name: str
    metadata: dict[str, Any] = {}


class ClassificationComponent(Component):
    value: tuple[str, float]
    type: Literal["classification"] = "classification"


def get_section_title(paragraph: Paragraph) -> str:
    if not paragraph.sections:
        return "Unknown section"
    return paragraph.sections[0].title


def parse_quotes(s: str) -> list[str]:
    if s.strip() == "n/a":
        return []
    return [clean_quote for line in s.split("\n") if (clean_quote := line.strip('" '))]


def format_paragraphs(paragraphs: list[Paragraph]) -> str:
    return "\n".join(f'"{paragraph}"' for paragraph in paragraphs)


def format_list(xs: list[str]) -> str:
    return "\n".join(f"- {x}" for x in xs)


def make_paragraph_placebo_classification_prompt(paragraph: Paragraph) -> str:
    return f"""Each of the paragraphs below is from a different paper. For each paragraph, classify whether a placebo was used in the paper.

Relevant definitions:
{DEFINITIONS_STRING}

###

Below is a paragraph from the "Methods" section of scientific paper A.

Paragraph from paper A: "Between the 26th and 29th of April 1998, all households in the intervention villages received a number of ITNs (insecticide-treated bed nets), according to the number of family members and were instructed explicitly about the correct use of the nets (Figure 2). More than 5000 ITN were distributed. The ITN (green colour, polyester, size 130 × 180 × 150 cm (11.6 m2 ) or 190 × 180 × 150 cm (14.5 m2 ), Siam-Dutch Co, Thailand) were already impregnated with deltamethrin (25 mg/m2 ) by the manufacturer. All households of the control villages received ITNs after the study was completed in May 1999."

Question: Based on the paragraph from paper A, did the paper use a placebo? Give your reasoning step by step, then answer with "Answer: Yes", "Answer: No", or "Answer: Unclear".

Reasoning: The paragraph from paper A doesn't explicitly talk about placebos. It's unclear whether bed nets that were not treated with insecticice were used as a placebo.

Answer: Unclear

###

Below is a paragraph from the "Randomization" section of scientific paper B.

Paragraph from paper B: To facilitate community-based treatment of malaria with the assigned regimen (artemether-lumefantrine or dihydroartemisinin-piperaquine), and to ensure that children received the correct regimen if they attended at health centres in the study area, ID cards were colour-coded according to intervention group and labelled with the regimen to be used for case management. The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded).

Question: Based on the paragraph from paper B, did the paper use a placebo? Give your reasoning step by step, then answer with "Answer: Yes", "Answer: No", or "Answer: Unclear".

Reasoning: The paragraph from paper B mentions that there was a placebo as alternative to seasonal malaria chemoprevention.

Answer: Yes

###

Below is a paragraph from the "Introduction" section of scientific paper C.

Paragraph from paper C: After a census of the village population, subjects in the target age group were screened for inclusion and exclusion criteria. Subjects who met inclusion and exclusion criteria were randomized either to receive two intermittent preventive treatments with standard recommended treatment doses of SP or no intermittent preventive treatment. Randomization codes were computer generated using simple randomization technique and treatment allocations were provided within sealed opaque envelopes.

Question: Based on the paragraph from paper C, did the paper use a placebo? Give your reasoning step by step, then answer with "Answer: Yes", "Answer: No", or "Answer: Unclear".

Reasoning: The paragraph from paper C says that the possible treatments were "two intermittent preventive treatments with standard recommended treatment doses of SP" and "no intermittent preventive treatment". The first group is the active treatment group. The second group seems to have no treatment, not a placebo treatment, so probably no placebo was used.

Answer: No

###

Below is a paragraph from the "Introduction" section of scientific paper D.

Paragraph from paper D: In India, the Integrated Child Development Service (ICDS) maintains a network of child-care centres, caring for children up to age 6 years and off ering the potential to deliver simple health interventions.

Question: Based on the paragraph from paper D, did the paper use a placebo? Give your reasoning step by step, then answer with "Answer: Yes", "Answer: No", or "Answer: Unclear".

Reasoning: The paragraph from paper D doesn't mention anything about placebos.

Answer: Unclear

###

Below is a paragraph from the "{get_section_title(paragraph)}" section of scientific paper E.

Paragraph from paper E: {paragraph}

Question: Based on the paragraph from paper E, did the paper use a placebo? Give your reasoning step by step, then answer with "Answer: Yes", "Answer: No", or "Answer: Unclear".

Answer:"""


def make_placebo_classification_from_arms_prompt(
    paper: Paper, arms: list[str], descriptions: list[str]
) -> str:
    arms_str = ""
    for i, (arm, description) in enumerate(zip(arms, descriptions)):
        arms_str += f"Arm {i+1} of paper D: {arm}\n"
        arms_str += f"Description of arm {i+1} of paper D: {description}\n\n"
    prompt = f"""For each paper, classify whether the paper used a placebo or not.

Take into account that a Placebo-controlled study is a way of testing a treatment in which a separate control group receives a sham "placebo" treatment which is specifically designed to be indistinguishable from the real treatment but has no real effect.

An open-label trial, or open trial is a type of clinical trial in which information is not withheld from trial participants. In particular, both the researchers and participants know which treatment is being administered.

Err on the side of caution: If you are unsure, answer "Unclear".

###

Paper A

First paragraph of paper A: In this study we investigate the effects of a single dose of oral zinc, and of oral calcium carbonate, on the risk of developing a type of complex polyomaviruses (CPV) that is associated with a high risk of death.

Arm 1 of paper A: Oral zinc supplement

Arm 2 of paper A: Oral calcium carbonate

Arm 3 of paper A: Placebo for oral zinc supplement

Question: Did paper A use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Reasoning: Arm 3 of paper A explicitly states that it is a placebo, so the answer is "Yes".

Answer: Yes

###

Paper B

First paragraph of paper B: It is suggested that a low intake of fish and/or n-3 PUFA is associated with depressed mood. However, results from epidemiologic studies are mixed, and randomized trials have mainly been performed in depressed patients, yielding conflicting results.

Arm 1 of paper B: 1800 mg/d EPA+DHA
Description of arm 1 of paper B: The participants received a capsule.

Arm 2 of paper B: No treatment
Description of arm 2 of paper B: The participants did not receive anything.

Question: Did paper B use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Reasoning: The description of arm 2 of paper B says that the participants did not receive anything, which implies that they also did not receive a placebo, so the answer is that paper B did not use a placebo.

Answer: No

###

Paper C

First paragraph of paper C: OBJECTIVE To assess the impact of zinc supplementation on nutritional and biochemical parameters among children aged 12 to 59 months. METHODS
A blinded randomized clinical trial was carried out with 58 children aged 12 to 59 months included in the Programa Governamental de Combate a Carências Nutricionais (National Child Nutritional Program), which provided them with 2 kg of iron-fortified milk. The supplementation group (n = 28) received 10 mg/day of zinc sulfate for four months, and the control group (n = 30) received placebo.

Arm 1 of paper C: Control group
Description of arm 1 of paper C: n = 30 participants

Arm 2 of paper C: Supplementation group (n = 28)
Description of arm 2 of paper C: Receiving 10 mg/day of zinc sulfate

Question: Did paper C use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Reasoning: Arm 1 of paper C does not include any information besides the number of participants and the fact that it is a control group, so we don't know if it is a placebo arm or not.

Answer: Unclear

###

Paper D

First paragraph of paper D: {paper.nonempty_paragraphs()[0]}

{arms_str}Question: Did paper D use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Answer:"""
    return prompt


def make_placebo_arm_index_prompt(
    paper: Paper, arms: list[str], descriptions: list[str]
) -> tuple[str, list[str]]:
    arms_str = ""
    for i, (arm, description) in enumerate(zip(arms, descriptions)):
        arms_str += f"Arm {i+1} of paper D: {arm}\n"
        arms_str += f"Description of arm {i+1} of paper D: {description}\n\n"

    choices = [str(i + 1) for i in range(len(arms))] + ["Unclear"]
    choices_str = ", ".join([f'"{i}"' for i in choices])

    prompt = f"""For each paper, find out which experimental arm of the paper was a placebo.

###

Paper A

First paragraph of paper A: In this study we investigate the effects of a single dose of oral zinc, and of oral calcium carbonate, on the risk of developing a type of complex polyomaviruses (CPV) that is associated with a high risk of death.

Arm 1 of paper A: Oral zinc supplement

Arm 2 of paper A: Oral calcium carbonate with zinc placebo

Arm 3 of paper A: Placebo for oral zinc supplement

Question: Did paper A use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Answer: Yes

Question: Which arm(s) of paper A used a placebo? Answer with one or more of "1", "2", "3", "Unclear".

Answer: 2, 3

###

Paper D

First paragraph of paper D: {paper.nonempty_paragraphs()[0]}

{arms_str}Question: Did paper D use a placebo? Give your reasoning, then answer with "Yes", "No", or "Unclear".

Answer: Yes

Question: Which arm(s) of paper D used a placebo? Answer with one or more of {choices_str}.

Answer:"""
    return prompt, choices


class PlaceboTree(Recipe):
    do_not_test = True

    async def run(self, *, paper: Paper):
        arms_recipe = ExperimentArms(mode=self.mode)
        arms, arm_descriptions = await arms_recipe.run(paper=paper)  # type: ignore[call-arg]

        placebo_class, placebo_description_draft = await self.classify_placebo(
            paper, arms, arm_descriptions
        )

        if placebo_class == "Placebo":
            placebo_description = await self.describe_placebo(
                paper, placebo_description_draft
            )
        else:
            placebo_description = placebo_class

        # Save results
        experiments: list[str] = list_experiments(
            document_id=get_full_document_id(paper.document_id)
        )
        for experiment in experiments:
            coarse_placebo_class = (
                "Placebo"
                if placebo_class == "Placebo"
                else "No placebo or placebo not mentioned"
            )
            recipe_result = RecipeResult(
                document_id=paper.document_id,
                question_short_name="placebo",
                experiment=experiment,
                classifications=[
                    placebo_class,
                    coarse_placebo_class,
                ],
                answer=placebo_description,
                result=placebo_description,
                excerpts=[],
            )
            self.maybe_add_to_results([recipe_result])

        return placebo_class, placebo_description

    async def classify_placebo(
        self, paper: Paper, arms: list[str], arm_descriptions: list[str]
    ) -> tuple[str, str | None]:
        """
        Classify whether the paper used a placebo by combining the results from
        classifying the arms and the paragraphs.
        """
        arm_cls, placebo_arm_indices = await self.classify_placebo_from_arms(
            paper, arms, arm_descriptions
        )

        if arm_cls == "Placebo":
            placebo_description_draft = "\n".join(
                f"{arms[i]}: {arm_descriptions[i]}" for i in placebo_arm_indices
            )
            return arm_cls, placebo_description_draft

        par_cls, par_prob = await self.classify_placebo_from_paragraphs(paper)
        combined_cls = self.combine_placebo_classifications(par_cls, par_prob, arm_cls)
        return combined_cls, None

    async def classify_placebo_from_arms(
        self, paper: Paper, arms: list[str], descriptions: list[str]
    ) -> tuple[str, list[int]]:
        """
        Classify whether the paper used a placebo by asking two questions:
        - Does the paper say that a placebo was used in the experiment?
        - If yes, which arm(s) received the placebo and were they indistinguishable from the other groups?
        """
        prompt = make_placebo_classification_from_arms_prompt(paper, arms, descriptions)
        classification = await self._classify(
            name="placebo_classification_from_arms",
            prompt=prompt,
            choices=[" Yes", " No", " Unclear"],
        )
        response_str = classification.value[0]
        if response_str.strip() == "No":
            return "No placebo", []
        elif response_str.strip() == "Unclear":
            return "Unclear", []

        assert response_str.strip() == "Yes", response_str

        arm_index_prompt, choices = make_placebo_arm_index_prompt(
            paper, arms, descriptions
        )

        arm = await self._classify(
            name="placebo_classification_from_arms_index",
            prompt=arm_index_prompt,
            choices=[f" {choice}" for choice in choices],
        )

        try:
            arm_indices = [int(x) - 1 for x in arm.value[0].strip().split(",")]
        except ValueError:
            return "Unclear", []

        distinguishability_prompt = f"""{arm_index_prompt} {arm.value[0]}

Followup-question: Are the placebo arms indistinguishable from the other groups for the participants? Give your reasoning, then answer with "Indistinguishable" or "Not indistinguishable".

Answer:"""

        distinguishability = await self._classify(
            name="placebo_classification_from_arms_distinguishable",
            prompt=distinguishability_prompt,
            choices=[" Indistinguishable", " Not indistinguishable"],
        )
        distinguishability_str = distinguishability.value[0]
        if distinguishability_str.strip() == "Indistinguishable":
            return "Placebo", arm_indices
        else:
            return "Unclear", []

    async def classify_placebo_from_paragraphs(
        self,
        paper: Paper,
    ) -> tuple[str, float]:
        """
        Classify whether the paper used a placebo by classifying each paragraph
        and aggregating the results.
        """
        paragraph_classifications = await self.classify_all_paragraphs(paper)
        return self.aggregate_paragraph_placebo_classifications(
            paragraph_classifications=paragraph_classifications,
        )

    async def classify_all_paragraphs(
        self, paper: Paper
    ) -> list[ClassificationComponent]:
        """
        Classify each non-empty paragraph of the paper as either "Yes", "No", or "Unclear"
        for whether it says that a placebo was used in the experiment.
        """
        paragraph_classifications = await map_async(
            paper.nonempty_paragraphs(),
            lambda paragraph: self.classify_paragraph_placebo(paragraph),
            max_concurrency=self.max_concurrency(),
            show_progress_bar=True,
        )
        return paragraph_classifications

    async def classify_paragraph_placebo(
        self, paragraph: Paragraph
    ) -> ClassificationComponent:
        """
        Classify a single paragraph as either "Yes", "No", or "Unclear"
        for whether it says that a placebo was used in the experiment.
        """
        prompt = make_paragraph_placebo_classification_prompt(paragraph)
        component = await self._classify(
            name="placebo_classification",
            prompt=prompt,
            choices=["Yes", "No", "Unclear"],
        )
        cls, prob = component.value
        if cls != "Yes":
            return component
        if "placebo" in str(paragraph).lower():
            return ClassificationComponent(
                name="placebo_classification",
                value=("Yes", 1.0),
                metadata={
                    "explanations": (component.metadata["explanations"] or "")
                    + "\nExplicitly mentions placebo"
                },
            )
        return ClassificationComponent(
            name="placebo_classification",
            value=("Yes", 0.5),
            metadata={
                "explanations": (component.metadata["explanations"] or "")
                + "\nImplicitly mentions placebo"
            },
        )

    def aggregate_paragraph_placebo_classifications(
        self,
        paragraph_classifications: list[ClassificationComponent],
    ) -> tuple[str, float]:
        """
        Aggregate the paragraph-wise placebo classification results
        into an overall classification of whether the paper used a
        placebo.
        """

        def class_prob(clf: ClassificationComponent) -> float:
            cls, prob = clf.value
            return prob

        def is_unclear(clf: ClassificationComponent) -> bool:
            cls, prob = clf.value
            return cls == "Unclear"

        pros = []
        cons = []
        for placebo_classification in paragraph_classifications:
            cls, _ = placebo_classification.value
            if cls == "Yes":
                pros.append(placebo_classification)
            elif cls == "No":
                cons.append(placebo_classification)
        if pros and not cons:
            answer = "Placebo"
            prob = max(class_prob(clf) for clf in pros)
        elif cons and not pros:
            answer = "No placebo"
            prob = 0.5
        elif not pros and not cons:
            answer = "Not mentioned"
            prob = 0.5
        else:  # pros and cons
            pprint([clf for clf in paragraph_classifications if not is_unclear(clf)])
            answer = "Unclear"
            prob = 0.5
        return answer, prob

    def combine_placebo_classifications(
        self, par_cls: str, par_prob: float, arm_cls: str
    ) -> str:
        # TODO: Be more systematic about probability aggregation
        if par_prob == 1.0:
            return par_cls
        match (par_cls, arm_cls):
            case ("Placebo", "Unclear"):
                return "Placebo"
            case ("Unclear", "Placebo"):
                return "Placebo"
            case ("Not mentioned", "Placebo"):
                return "Placebo"
            case ("Placebo", "No placebo"):
                return "Unclear"
            case ("No placebo", "Placebo"):
                return "Unclear"
            case ("Unclear", "No placebo"):
                return "Unclear"
            case ("Not mentioned", "Unclear"):
                return "Not mentioned"
            case ("No placebo", "Unclear"):
                return "No placebo"
            case ("Not mentioned", "No placebo"):
                return "No placebo"
            case _:
                assert par_cls == arm_cls
                return par_cls

    async def describe_placebo(
        self, paper: Paper, description_draft: str | None
    ) -> str:
        """
        Describe the placebo or placebos used in the study, either by using the draft
        from the arm classification or by asking a QA question.
        """

        if description_draft is not None:
            return description_draft

        qa_recipe = ComparisonsQA(mode=self.mode)
        qa_result = await qa_recipe.run(
            paper=paper,
            question_short="""What was the placebo?""",
            question_long="""What was the placebo, or what were the placebos used in the study? Describe in a few sentences.""",
            answer_prefix="Answer:",
            num_paragraphs=3,
        )
        return qa_result

    async def _classify(
        self,
        *,
        name: str,
        prompt: str,
        choices: list[str],
        unknown_choice: str = "Unclear",
        unknown_threshold: float = 0.8,
        verbose: bool = False,
        agent_name: str = "instruct-reasoning-crowd",
    ) -> ClassificationComponent:
        answer_probs, explanations = await self.agent(
            agent_name="instruct-reasoning-crowd"
        ).classify(
            prompt=prompt,
            choices=tuple(choices),
            verbose=verbose,
        )
        answer, probability = max_by_value(answer_probs)
        if probability < unknown_threshold:
            answer = unknown_choice
            probability = 0.9
        return ClassificationComponent(
            name=name,
            value=(answer, probability),
            metadata={"explanations": explanations},
        )



from collections.abc import Sequence

from structlog.stdlib import get_logger

from ice.apis.openai import openai_complete
from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.value import numbered_list
from ice.recipe import recipe
from ice.recipes.program_search.nodes.answer.generate_reasoning.prompts import (
    add_reasoning,
)
from ice.recipes.program_search.nodes.answer.types import Demonstration
from ice.recipes.program_search.nodes.answer.types import DemonstrationWithReasoning

log = get_logger()

INSTRUCTIONS = "Answer the question based on the excerpts provided."
FEW_SHOT_INSTRUCTIONS = (
    "Answer each question based on the excerpts provided for that question."
)
REASONING_ADDITIONAL = " Be sure to include your reasoning, then your final answer."


async def simple_answer(question: str, texts: Sequence[str]):
    prompt = "\n\n".join(
        (
            INSTRUCTIONS,
            f"Question: {question}",
            "Excerpts:",
            numbered_list(texts).transform(),
            f"""Answer to the question "{question}":""",
        )
    )
    return await recipe.agent().complete(prompt=prompt)


DEMONSTRATION_EXAMPLE = """
Question: {question}

Excerpts:

{texts}

Answer to the question "{question}": {answer}
""".strip()


def demonstration_answer_prompt(
    question: str, texts: Sequence[str], demonstrations: Sequence[Demonstration]
) -> str:
    examples = format_multi(
        DEMONSTRATION_EXAMPLE,
        [d.as_dict() for d in demonstrations]
        + [dict(question=question, texts=numbered_list(texts), answer=stop(""))],
    )
    return "\n\n".join([FEW_SHOT_INSTRUCTIONS, "\n\n---\n\n".join(examples)])


DEMONSTRATION_WITH_REASONING_EXAMPLE = """
Question: {question}

Excerpts:

{texts}

Answer to the question "{question}".

Let's think it over: First, Excerpt 1 says that {reasoning}

Final answer:

{answer}
""".strip()


def demonstration_with_reasoning_answer_prompt(
    question: str,
    texts: Sequence[str],
    demonstrations: Sequence[DemonstrationWithReasoning],
):
    examples = format_multi(
        DEMONSTRATION_WITH_REASONING_EXAMPLE,
        [d.as_dict() for d in demonstrations]
        + [dict(question=question, texts=numbered_list(texts), reasoning=stop(""))],
    )
    return "\n\n".join(
        [FEW_SHOT_INSTRUCTIONS + REASONING_ADDITIONAL, "\n\n---\n\n".join(examples)]
    )


SUPPRESS_EOT = {"50256": -100}


async def demonstration_answer(
    question: str, texts: Sequence[str], demonstrations: Sequence[Demonstration]
):
    prompt = demonstration_answer_prompt(question, texts, demonstrations)
    return await recipe.agent().complete(prompt=prompt, stop="\n\n---")


async def _get_reasoning(initial_prompt: str, completion: str):
    try:
        return completion.split("Final answer:")[1].strip()
    except IndexError:
        lines = completion.splitlines()
        if not len(lines) == 1:
            raise ValueError("Unexpected response")
        new_prompt = initial_prompt + " " + lines[0].strip() + "\n\n" + "Final answer:"
        new_completion = await openai_complete(
            prompt=new_prompt, stop="\n\n---", max_tokens=200
        )
        log.warning(
            "Final answer not included in initial response",
            initial_response=completion,
            initial_prompt=initial_prompt,
            new_prompt=new_prompt,
            new_final_answer=new_completion,
        )
    return new_completion["choices"][0]["text"].strip()


async def demonstration_answer_with_reasoning(
    question: str, texts: Sequence[str], demonstrations: Sequence[Demonstration]
):
    with_reasoning = await add_reasoning(demonstrations)
    prompt = demonstration_with_reasoning_answer_prompt(
        question=question, texts=texts, demonstrations=with_reasoning
    )
    max_tokens = 500
    while True:
        completion = await openai_complete(
            prompt=prompt,
            stop="\n\n---",
            max_tokens=max_tokens,
        )
        try:
            answer: str = completion["choices"][0]["text"]
            break
        except IndexError:
            finish_reason = completion["choices"][0]["finish_reason"]
            if finish_reason == "length" and max_tokens < 1_500:
                max_tokens += 200
            else:
                raise

    return await _get_reasoning(prompt, answer)


async def elicit_answer_prompt(question: str, text: str) -> str:
    prompt = f"""Answer the question "{question}" based on the excerpt from a research paper. \
Include everything that the paper excerpt has to say about the answer. \
Make sure everything you say is supported by the excerpt. \
The excerpt may cite other papers; \
answer about the paper you're reading the excerpt from, not the papers that it cites. \
Answer in one phrase or sentence:

Paper excerpt: {text}

Question: {question}

Answer:"""
    completion = await openai_complete(
        prompt=prompt,
        stop=None,
        max_tokens=200,
    )

    return completion["choices"][0]["text"]


recipe.main(simple_answer)



import pandas as pd
import pytest

from ice import trace


@pytest.mark.anyio
async def test_trace_blocks():
    trace.enable_trace()
    current_trace = trace.trace_var.get()
    assert trace.trace_enabled()
    assert current_trace

    long = "a" * trace.Trace.BLOCK_LENGTH
    assert trace.emit_block("foo") == (0, 0)
    assert trace.emit_block("bar") == (0, 1)
    assert trace.emit_block(long) == (0, 2)
    assert trace.emit_block("baz") == (1, 0)
    assert trace.emit_block("quux") == (1, 1)

    # Now repeat the same blocks and check caching
    assert trace.emit_block("baz") == (1, 0)
    assert trace.emit_block("bar") == (0, 1)
    assert trace.emit_block(long) == (0, 2)
    assert trace.emit_block("foo") == (0, 0)
    assert trace.emit_block("foo") == (0, 0)
    assert trace.emit_block("foo") == (0, 0)
    assert trace.emit_block(long) == (0, 2)
    assert trace.emit_block("quux") == (1, 1)

    assert (
        current_trace.dir / "block_0.jsonl"
    ).read_text() == f'"foo"\n"bar"\n"{long}"\nend\n'


def test_get_strings():
    assert trace.get_strings("foo") == ["foo"]
    assert trace.get_strings(["foo", "bar"]) == ["foo", "bar"]
    assert trace.get_strings({"foo": "bar"}) == ["bar"]

    # First descendant list of strings
    assert trace.get_strings({"foo": {"spam": ["bar", "baz"], "x": "y"}}) == [
        "bar",
        "baz",
    ]

    # Empty values that get converted to '()'
    assert trace.get_strings([]) == ["()"]
    assert trace.get_strings({}) == ["()"]
    assert trace.get_strings(None) == ["()"]
    assert trace.get_strings("") == ["()"]

    # Other falsy values are just their string representation
    assert trace.get_strings(0) == ["0"]
    assert trace.get_strings(0.0) == ["0.0"]
    assert trace.get_strings(False) == ["False"]

    # First descendant isn't a string, gets converted
    assert trace.get_strings([1, 2, 3]) == ["1"]

    # Long lists get truncated
    assert trace.get_strings(["1", "2", "3", "4", "5"]) == ["1", "2", "3", "..."]

    # Long strings get truncated
    assert trace.get_strings("a" * 100) == ["a" * 35 + "..."]

    # `value` gets extracted
    assert trace.get_strings({"a": "b", "c": "d"}) == ["b"]
    assert trace.get_strings({"value": "b", "c": "d"}) == ["b"]
    assert trace.get_strings({"a": "b", "value": "d"}) == ["d"]
    assert trace.get_strings({"a": "b", "value": {}}) == ["()"]

    # self and record are omitted
    assert trace.get_strings({"self": "foo", "record": "bar", "a": "b", "c": "d"}) == [
        "b"
    ]
    assert trace.get_strings({"self": "foo", "record": "bar"}) == ["()"]
    assert trace.get_strings(
        {
            "x": "y",
            "value": {"self": "foo", "record": "bar", "a": "b", "c": "d"},
        }
    ) == ["b"]

    # Non-strings are filtered out
    assert trace.get_strings(["a", "b", 3, None, "c"]) == ["a", "b", "c"]

    # Nested lists are not flattened
    assert trace.get_strings([["a", "b"], ["c", "d"]]) == ["a", "b"]

    assert trace.get_strings(pd.DataFrame({"a": [1, 2, 3]})) == [
        """\
   a
0  1
1  2
2  3"""
    ]



from collections.abc import Sequence

from ice.paper import Paper
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.prompts.can_count_exps import (
    CAN_WE_COUNT_EXPERIMENTS_BEST_CHOICE,
)
from ice.recipes.experiments_and_arms.prompts.can_count_exps import (
    CAN_WE_COUNT_EXPERIMENTS_CHOICES,
)
from ice.recipes.experiments_and_arms.prompts.can_count_exps import (
    CAN_WE_COUNT_EXPERIMENTS_REASONING_STOP,
)
from ice.recipes.experiments_and_arms.prompts.can_count_exps import get_helpfulness
from ice.recipes.experiments_and_arms.prompts.can_count_exps import get_reasoning
from ice.recipes.experiments_and_arms.prompts.can_count_exps import (
    make_can_we_count_experiments_prompt,
)
from ice.recipes.experiments_and_arms.prompts.count_exps import (
    COUNT_EXPERIMENTS_REASONING_STOP,
)
from ice.recipes.experiments_and_arms.prompts.count_exps import count_from_answer
from ice.recipes.experiments_and_arms.prompts.count_exps import get_count_exps_reasoning
from ice.recipes.experiments_and_arms.prompts.count_exps import (
    make_count_experiments_prompt_func,
)
from ice.recipes.experiments_and_arms.prompts.passages_to_keep import (
    keep_most_helpful_paragraphs,
)
from ice.recipes.experiments_and_arms.prompts.utils import plurality_greedy
from ice.recipes.experiments_and_arms.recipes.best_passages import (
    rate_helpfulness_with_reasoning,
)
from ice.recipes.experiments_and_arms.recipes.reason_select_and_answer import (
    answer_with_best_reasoning,
)
from ice.recipes.experiments_and_arms.types import PassageWithReasoning


async def count_experiments(
    paper: Paper,
) -> tuple[PassageWithReasoning[int], Sequence[str]]:
    """How many distinct experiments are in the paper?

    Args:
        paper (Paper): Paper to count experiments in.

    Returns:
        tuple[PassageWithReasoning[int], Sequence[str]]: The number of experiments and the paragraphs used to count the experiments.
    """
    paragraphs = paper.nonempty_paragraphs()
    passages_by_relevance = await rate_helpfulness_with_reasoning(
        [str(p) for p in paragraphs],
        make_can_we_count_experiments_prompt,
        CAN_WE_COUNT_EXPERIMENTS_CHOICES,
        CAN_WE_COUNT_EXPERIMENTS_BEST_CHOICE,
        reasoning_stop=CAN_WE_COUNT_EXPERIMENTS_REASONING_STOP,
        get_reasoning=get_reasoning,
        get_helpfulness=get_helpfulness,
        num_shots=1,
        passages_per_prompt=4,
        step=1,
    )
    paragraphs_to_keep = await keep_most_helpful_paragraphs(passages_by_relevance)

    experiment_count = await answer_with_best_reasoning(
        10,
        plurality_greedy,
        paragraphs_to_keep,
        2,
        0.4,
        COUNT_EXPERIMENTS_REASONING_STOP,
        make_count_experiments_prompt_func,
        get_count_exps_reasoning,
        None,
        lambda resp: count_from_answer(resp["choices"][0]["text"]),
    )

    return experiment_count, paragraphs_to_keep


recipe.main(count_experiments)



from ice.apis.openai import openai_complete
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.prompts.quick_evaluate import get_grade
from ice.recipes.experiments_and_arms.prompts.quick_evaluate import (
    make_quick_eval_prompt,
)
from ice.recipes.experiments_and_arms.types import ExperimentsArms


async def quick_evaluate(
    gs: ExperimentsArms, generated: ExperimentsArms
) -> tuple[str, str]:
    """Quickly evaluate the generated experiments and arms against the gold standard.

    Args:
        gs (ExperimentsArms): The gold standard.
        generated (ExperimentsArms): The generated answer.

    Returns:
        tuple[str, str]: The evaluation and the letter grade.
    """
    prompt, stop_seq = make_quick_eval_prompt(gs, generated)
    response = (await openai_complete(prompt, stop=stop_seq))["choices"][0]["text"]
    return response, get_grade(response)


recipe.main(quick_evaluate)



from collections.abc import Sequence

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform

QUESTION_FREE_EXAMPLES: list[dict[str, str | StopSentinel]] = [
    dict(
        context="Lisa loves to play practical jokes.",
        passage="But sometimes she goes too far.",
        rewrite="But sometimes she [Lisa] goes too far.",
    ),
    dict(
        context="The Super Bowl XLI halftime show took place on February 4, 2007.",
        passage="It was headlined by Prince.",
        rewrite="It [The Super Bowl XLI halftime show] was headlined by Prince.",
    ),
    dict(
        context="More than one fifth of the world’s population lives on less than Purchasing Power Parity (PPP) US$1.25 a day, and there is an emerging international consensus that this share should (and can) be driven close to zero by 2030 (1, 2).",
        passage="Reaching this objective will require enabling the poorest families, who are often the most marginalized within their villages, to shift from insecure and fragile sources of income to more sustainable income-generating activities.",
        rewrite="Reaching this objective [driving the share of the world’s population that lives on less than Purchaing Power Parity (PPP) US$1.25 a day from more than one fifth of the to zero by 2030] will require enabling the poorest families, who are often the most marginalized within their villages, to shift from insecure and fragile sources of income to more sustainable income-generating activities.",
    ),
    dict(
        context='We present results from randomized control trials (RCTs) in six countries of a particular approach to foster self-employment activities amongst the very poor. Originally designed and implemented by BRAC, a large Bangladeshi NGO that runs several country-wide programs, the “Graduation” program provides a holistic set of services, including the grant of a productive asset, to the poorest households in a village (referred to by BRAC as the “ultra-poor”). The beneficiaries [of the Graduation program, the poorest housholds in a village, or the "ultra-poor"] are identified through a participatory process in a village meeting, followed by a verification visit by the organization’s [the implmenter of the "Graduation" program] staff. Selected beneficiaries [among the poorest housholds in a village, or the "ultra-poor"] are then given a productive asset [by the implementer of the "Graduation Program"] that they choose from a list, training and support for the asset they have chosen, as well as general life skills coaching, weekly consumption support for some fixed period, and typically access to savings accounts and health information or services.',
        passage="These different activities (plus regular interactions with the households over the course of a year) are designed to complement each other in helping households to start a productive self-employment activity.",
        rewrite='These different activities [training and support for the assest they have chosen and received, general life skills coaching, weekly consumption support, and typically access to savings accounts and health information or services] (plus regular interactions with the households over the course of a year) are designed to complement each other in helping households [beneficiaries selected for the "Graduation" program from among the poorest housholds in a village, or the "ultra-poor"] to start a productive self-employment activity.',
    ),
]

QUESTION_GUIDED_EXAMPLES: list[
    dict[str, str | ValueTransform[Sequence[str]] | StopSentinel]
] = [
    dict(
        questions=numbered_list(["What were Prince's biggest concerts?"]),
        context="The Super Bowl XLI halftime show took place on February 4, 2007.",
        passage="It was headlined by Prince.",
        rewrite="It [The Super Bowl XLI halftime show] was headlined by Prince.",
    ),
    dict(
        questions=numbered_list(["What were the aims of the study?"]),
        context="More than one fifth of the world’s population lives on less than Purchasing Power Parity (PPP) US$1.25 a day, and there is an emerging international consensus that this share should (and can) be driven close to zero by 2030 (1, 2).",
        passage="Reaching this objective will require enabling the poorest families, who are often the most marginalized within their villages, to shift from insecure and fragile sources of income to more sustainable income-generating activities.",
        rewrite="Reaching this objective [driving the share of the world’s population that lives on less than Purchaing Power Parity (PPP) US$1.25 a day from more than one fifth of the to zero by 2030] will require enabling the poorest families, who are often the most marginalized within their villages, to shift from insecure and fragile sources of income to more sustainable income-generating activities.",
    ),
    dict(
        questions=numbered_list(
            ["Who funded the study?", "What were the limitations of the RCT findings?"]
        ),
        context='We present results from randomized control trials (RCTs) in six countries of a particular approach to foster self-employment activities amongst the very poor. Originally designed and implemented by BRAC, a large Bangladeshi NGO that runs several country-wide programs, the “Graduation” program provides a holistic set of services, including the grant of a productive asset, to the poorest households in a village (referred to by BRAC as the “ultra-poor”). The beneficiaries [of the Graduation program, the poorest housholds in a village, or the "ultra-poor"] are identified through a participatory process in a village meeting, followed by a verification visit by the organization’s [the implmenter of the "Graduation" program] staff. Selected beneficiaries [among the poorest housholds in a village, or the "ultra-poor"] are then given a productive asset [by the implementer of the "Graduation Program"] that they choose from a list, training and support for the asset they have chosen, as well as general life skills coaching, weekly consumption support for some fixed period, and typically access to savings accounts and health information or services.',
        passage="These different activities (plus regular interactions with the households over the course of a year) are designed to complement each other in helping households to start a productive self-employment activity.",
        rewrite="These different activities (plus regular interactions with the households over the course of a year) are designed to complement each other in helping households to start a productive self-employment activity.",
    ),
]
QUESTION_FREE_PREFIX = "Instructions: Enrich each Passage with the Context."
QUESTION_GUIDED_PREFIX = "Instructions: Enrich each Passage with the Context, carrying forward all and only information that can help answer the questions."

QUESTION_FREE_EXAMPLE_TEMPLATE = """Context: {context}
Passage: {passage}
Rewrite: {rewrite}
""".strip()

QUESTION_GUIDED_EXAMPLE_TEMPLATE = """

Questions:

{questions}

Context: {context}
Passage: {passage}
Rewrite: {rewrite}""".strip()


def decontext_prompt(
    context: str, passage: str, questions: Sequence[str] | None = None
) -> str:
    last_example: dict[str, str | StopSentinel | ValueTransform] = dict(
        context=context, passage=passage, rewrite=stop("")
    )
    if questions:
        last_example["questions"] = numbered_list(questions)
    examples = format_multi(
        QUESTION_GUIDED_EXAMPLE_TEMPLATE
        if questions
        else QUESTION_FREE_EXAMPLE_TEMPLATE,
        QUESTION_GUIDED_EXAMPLES + [last_example]
        if questions
        else QUESTION_FREE_EXAMPLES + [last_example],
    )
    return "\n\n".join(
        (
            QUESTION_GUIDED_PREFIX if questions else QUESTION_FREE_PREFIX,
            "\n\n---\n\n".join(examples),
        )
    )



from importlib import import_module
from inspect import signature
from pathlib import Path

import pytest

from faker import Faker

from ice.paper import Paper
from ice.recipe import FunctionBasedRecipe
from ice.recipe import recipe
from ice.trace import enable_trace

Faker.seed(0)
fake = Faker()


root_dir = Path(__file__).parent.parent

recipe.all_recipes = []
primer_recipes_dir = root_dir / "ice" / "recipes" / "primer"
for path in primer_recipes_dir.glob("**/*.py"):
    relative_path = path.relative_to(primer_recipes_dir)
    module_name = ".".join(relative_path.parts[:-1] + (relative_path.stem,))
    import_module(f"ice.recipes.primer.{module_name}")

paper = Paper.load(Path(root_dir / "papers" / "keenan-2018-tiny.txt"))


@pytest.mark.parametrize("main", recipe.all_recipes)
@pytest.mark.anyio
async def test_all_primer_recipes(main: FunctionBasedRecipe):
    kwargs = {}
    for p in signature(main).parameters.values():
        if p.default is not p.empty:
            value = p.default
        elif issubclass(p.annotation, str):
            value = fake.sentence()
        elif issubclass(p.annotation, Paper):
            value = paper
        else:
            raise ValueError(f"Cannot handle parameter {p}")
        kwargs[p.name] = value
    enable_trace()
    await main(**kwargs)



from dataclasses import dataclass

from fvalues import F
from pydantic import BaseModel

from ice.json_value import to_json_value


@dataclass
class MyDataClass:
    things: list


class MyModel(BaseModel):
    stuff: dict


def test_objects():
    assert to_json_value(MyDataClass([1, 2, 3])) == {"things": [1, 2, 3]}
    assert to_json_value(MyModel(stuff={"a": 1})) == {"stuff": {"a": 1}}
    assert to_json_value(
        MyDataClass(
            [
                MyModel(
                    stuff={
                        "foo": MyDataClass(
                            ["a", "b"],
                        ),
                        "bar": 9,
                    }
                ),
                "c",
            ]
        )
    ) == {
        "things": [
            {
                "stuff": {
                    "foo": {
                        "things": [
                            "a",
                            "b",
                        ]
                    },
                    "bar": 9,
                }
            },
            "c",
        ]
    }


def test_non_str_keys():
    assert to_json_value({"a": 1, 2: 3}) == {"a": 1, "2": 3}


def test_iterables():
    assert to_json_value([1, 2, 3]) == [1, 2, 3]
    assert to_json_value((1, 2, 3)) == [1, 2, 3]
    # Sets are unordered, so use a singleton for testing
    assert to_json_value({1}) == [1]


def test_function():
    def foo():
        pass

    assert to_json_value(foo) == {
        "class_name": "function",
        "name": "foo",
    }


def test_primitives():
    things = [1, 2.3, None, "4", True, False]
    assert to_json_value(things) == things
    for thing in things:
        assert to_json_value(thing) is thing


def test_repr():
    class Foo:
        def dict(self, x):
            # This doesn't get used because it has a parameter,
            # which leads to raising a TypeError when .dict() is called.
            return {"a": x}

        def __repr__(self):
            return "Foo"

    assert to_json_value(Foo()) == "Foo"


def test_f():
    f = F(f"hello {1 + 2} world")
    assert to_json_value(f) == {
        "__fstring__": [
            "hello ",
            {"formatted": "3", "source": "1 + 2", "value": 3},
            " world",
        ]
    }
